import unittest
import os
import shutil
import collections
import inspect
from osgeo import gdalconst
import numpy as np

import geoio.dg
import tinytools as tt
import geoio
import dgsamples

class TestGeoioSpectralFileHandling(unittest.TestCase):
    """Testing for :
    communication between geoio and DGAComp
    """

    def setUp(self):
        # Ensure the previous tests closed down as they were supposed to by
        # removing all spectral files

        # Setup test img

        self.clean_dir = os.path.dirname(dgsamples.wv2_longmont_1k.ms)
        #self.clean_dir = "../data/imagefiles/053792616010_01/053792616010_01_P001_MUL/"
        print("Copying data")
        test_dir = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))
        self.test_dir = test_dir+os.path.sep+"mul_test"
        shutil.copytree(self.clean_dir,self.test_dir)
        self.test_img =  self.test_dir+os.path.sep+\
                         "14JUN20181517-M2AS-053792616010_01_P001.TIL"
        self.img = geoio.dg.DGImage(self.test_img)

    def tearDown(self):
        # Remove all sepctral files
        del self.img
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_create_at_sensor_rad_files(self):
        pre=(self.img.files.rad==None)
        self.img.create_at_sensor_rad_files()
        post=(self.img.files.rad!=None)
        self.assertTrue(pre & post)

    def test_create_toa_ref_files(self):
        pre=(self.img.files.toa==None)
        self.img.create_toa_ref_files()
        post=(self.img.files.toa!=None)
        self.assertTrue(pre & post)

    @unittest.skip("Need to get off IDL before this can be stable.")
    def test_create_dgacomp_ref_files(self):
        pre=(self.img.files.dgacomp==None)
        self.img.create_dgacomp_ref_files()
        post=(self.img.files.dgacomp!=None)
        self.assertTrue(pre & post)

    def test_DGImage_delete_rad_files(self):
        pre=(self.img.files.rad==None)
        self.img.create_at_sensor_rad_files()
        post=(self.img.files.rad!=None)
        self.img.delete_rad_files(test_only=False)
        gone=(self.img.files.rad==None)
        self.assertTrue(pre & post & gone)

    def test_DGImage_delete_toa_ref_files(self):
        pre=(self.img.files.toa==None)
        self.img.create_toa_ref_files()
        post=(self.img.files.toa!=None)
        self.img.delete_toa_ref_files(test_only=False)
        gone=(self.img.files.toa==None)
        self.assertTrue(pre & post & gone)

    @unittest.skip("Need to get off IDL before this can be stable.")
    def test_DGImage_delete_dgacomp_files(self):
        pre=(self.img.files.dgacomp==None)
        self.img.create_dgacomp_ref_files()
        post=(self.img.files.dgacomp!=None)
        self.img.delete_dgacomp_files(test_only=False)
        gone=(self.img.files.dgacomp==None)
        self.assertTrue(pre & post & gone)

    def test_DGImage_delete_all_spectral_files(self):
        pre=((self.img.files.toa==None) &
             (self.img.files.rad==None))
        self.img.create_toa_ref_files()
        self.img.create_at_sensor_rad_files()
        post=((self.img.files.toa!=None) &
              (self.img.files.rad!=None))
        self.img.delete_all_spectral_files(test_only=False)
        gone=((self.img.files.toa==None) &
              (self.img.files.rad==None))
        self.assertTrue(pre & post & gone)

if __name__ == '__main__':
    unittest.main()


import tinytools as tt
import numpy as np
import matplotlib.pyplot as plt

# Currently doesn't return figure handles to help with memory size issues on
# large images.  This could probably use a close look at ways to make it
# more efficient:
# https://pypi.python.org/pypi/ModestImage
# https://github.com/ipython/ipython/issues/1623/
# http://stackoverflow.com/questions/15345336/memory-leak-in-matplotlib-imshow

def imshow(data,stretch=[0.02,0.98],stretch_type='linear'):
    """Convenience method to do all the plotting gymnastics to get a resonable
    looking image plot.

    Input:
    data            numpy array in gdal band order - 3 dimensions (bands, x, y)
    stretch         stretch values on a scale of [0,1]
    stretch_type    type of stretch scale (only linear is curretly supported)
    """

    if len(data[:,0,0]) != 3:
        raise ValueError('This convenience function is only implemented ' \
                         'for three bands.  Use img.get_data(bands=...) to ' \
                         'retrieve specific data.')
                        # ToDo - This could also be speed up by indexing a
                        # single numpy array.

    # define stretch
    # Possibly useful code for additional stretches at:
    # http://scikit-image.org/docs/dev/api/skimage.exposure.html
    # also
    # http://scikit-image.org/docs/dev/auto_examples/plot_equalize.html
    if stretch_type == "linear":
        pass
    else:
        raise ValueError('The passed value of stretch is not implemented.')

    # Get the per-band scaled data
    data = tt.np_img.conv_to_bandslast(data)
    data = data.astype('float32')
    lims = np.percentile(data,(2,98),axis=(0,1))
    for x in xrange(len(data[0,0,:])):
        top = lims[:,x][1]
        bottom = lims[:,x][0]
        data[:,:,x] = (data[:,:,x]-bottom)/float(top-bottom)
    data = np.clip(data,0,1)

    plt.imshow(data,interpolation='nearest');
    plt.show(block=False)

    # ToDo: fix the handle return for later update and memory issues - see
    # comments above
    # return handle

def hist(data):
    """Convenience method to do all the plotting gymnastics to get a resonable
    looking histogram plot.

    Input: data - numpy array in gdal format - (bands, x, y)

    Returns:  matplotlib figure handle

    Adapted from:
    http://nbviewer.jupyter.org/github/HyperionAnalytics/PyDataNYC2014/blob/master/color_image_processing.ipynb
    """

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    plt.hold(True)
    for x in xrange(len(data[:,0,0])):
        counts, edges = np.histogram(data[x,:,:],bins=100)
        centers = [(edges[i]+edges[i+1])/2.0 for i,v in enumerate(edges[:-1])]
        ax1.plot(centers,counts)
    plt.hold(False)

    plt.show(block=False)

    # return fig



import imp as _imp
import logging

# Pull in main modules/classes
from base import GeoImage
from dg import DGImage
import constants

logger = logging.getLogger(__name__)

# Pull in optional modules
try:
    _imp.find_module('matplotlib')
    import plotting
except:
    logger.info('plotting is not available from geoio - matplotlib is missing.')
    pass



'''
High level geo file image access

Base GeoImage class is a wrapper around gdal providing easy access to
image metadata, file read/write, and i/o.  Higher level classes (such
as DGImage) can build on this class to provide data specific metadata and
spectral transformations.
'''

from __future__ import division

from osgeo import gdal, gdalconst, osr, ogr
import numpy as np
import os
import warnings
import collections
import textwrap
import tempfile
import logging
import math
from tzwhere import tzwhere
import tinytools as tt

# package import
import constants as const

# Module setup
gdal.UseExceptions()
ogr.UseExceptions()
logger = logging.getLogger(__name__)
# To get access to logging statements from the command line:
# import logging
# logging.basicConfig(level=logging.DEBUG) # or your desired level

class OverlapError(ValueError):
    '''Raise when the window does not overlap the image.  This can be
    caught and passed when the window is expected to not overlap in
    some cases.
    '''
    pass

class GeoImage(object):
    """
    Base image class providing high-level access to image data and metadata
    as well as methods to easily interact with the image and related
    vector files.

    Functionality is tested against .TIL, .VRT, and .TIF formats, but this
    base class should support any GDAL format.

    Parameters
    ----------
    file_in : str
        String describing a file on disk that is of a valid input format.
    derived_dir : str
        The location to store files created by the class.

    Attributes
    ----------
    files : tinytools.bunch.OrderedBunch
        A collection of the image files used in the object.  The GeoImage
        class populates:

        -derived_dir : The directory into which create files are stored.
        -dfile : The input image file.
        -dfile_tiles : The tiles of the virtual data set - if the data set
                       doesn't contain tiles, this should be equal to dfile.

        Inherited classes can override/extend this listing.
    meta : tinytools.bunch.OrderedBunch
        Summary metadata of the base image.
    shape : tuple
        shape of the image in gdal format (bands,x,y).
    resolutions : tuple
        length 2 tuple with resolutions of x and y image dimensions.
    """

    def __init__(self, file_in, derived_dir=None):
        """Initialize class with data and meta-data from file.  __init__
        class is in the class definition. """

        ## Search for files that are needed
        if not os.path.isfile(file_in):
            raise ValueError("The file that was passed in does not exist.")

        ## Start populating files dictionary for geoimage info
        # ... variables created here:
        # self.files_dict['fin']
        # self.files_dict['gdal_file_list']
        # self.files_dict['dfile']
        # self.files_dict['dfile_tiles']
        # self.files_dict['derived_dir']

        # Get the file name and full file directory
        ifile = os.path.abspath(file_in)
        fname = os.path.basename(ifile)
        fdir = os.path.dirname(ifile)
        flist = os.listdir(fdir)

        # Create files dictionary to populate - this will be bunched later
        #!# self.files_dict = {}
        self.files = tt.bunch.OrderedBunch({})

        # Set the place to store/retrive derived files (i.e. spectral data)
        if derived_dir:
            if not os.path.isdir(derived_dir):
                raise ValueError("The requested derived data directory does "
                                 "not exist.")
            if not os.access(os.path.join(derived_dir, ''), os.W_OK):
                raise ValueError("Write access is required for the requested "
                                 "location passed into derived_dir.")
            self.files.derived_dir = os.path.join(derived_dir, '')
        else:
            if os.access(fdir, os.W_OK):
                self.files.derived_dir = fdir
            else:
                self.files.derived_dir = fdir
                warnings.warn("The input file location is not writable.  "
                              "Derived file creation (i.e. spectral files) "
                              "will not be available. Either write permissions "
                              "need to be provided or the object can be "
                              "reinstantiated with a writable location passed "
                              "to the input variable dervied_store_dir.")

        ### Setup the dataset and subdataset variables
        (tmpfile,tmptiles)=self._get_file_and_tiles(ifile)

        #!# self.files_dict['dfile'] = tmpfile
        #!# self.files_dict['dfile_tiles'] = tmptiles
        self.files.dfile = tmpfile
        self.files.dfile_tiles = tmptiles

        # Create the files dictionary bunch
        #!#self.files = tt.bunch.OrderedBunch(self.files_dict)

        ## Populate geoimage info
        # ... variables created here:
        # self._fobj
        # self.meta_geoimg_dict
        # self.meta

        # Open the image in files.dfile
        self._fobj = self._get_gdal_obj(self.files.dfile,
                                        self.files.dfile_tiles)

        # Populate metadata info from gdal
        self._set_metadata()

    def _get_file_and_tiles(self, ifile):

        # If fname is a .til file then create .vrt
        if tt.files.filter(ifile, '*.TIL', case_sensitive=False):
            file_loc = ifile
            tiles_loc = tt.pvl.read_from_pvl(ifile,'filename')
            dname = os.path.dirname(ifile)
            tiles_loc = [os.path.join(dname,x) for x in tiles_loc]

        # If fname is a VRT, pull subdatasets from the file object
        elif tt.files.filter(ifile, '*.VRT', case_sensitive=False):
            file_loc = ifile
            tmp = gdal.Open(file_loc) # Open to pull VRT file list
            tmp_files = tmp.GetFileList()
            tiles_loc = [x for x in tmp_files if not
                            tt.files.filter(x, '*.VRT', case_sensitive=False)]
            tmp = None # Close the opened file from above

        # If this is an ENVI file, then a file without an extension should
        # exist and will be the root file.
        elif os.path.isfile(os.path.splitext(ifile)[0]):
            tmp = os.path.splitext(ifile)[0]
            file_loc = tmp
            tiles_loc = [tmp]

        # If file input isn't a .TIL, .VRT, or ENVI file, just pass it into the
        # object vars to pass to gdal.
        else:
            # Else, just pass it on to open in gdal
            file_loc = ifile
            tiles_loc = [ifile]

        return (file_loc,tiles_loc)

    def _get_gdal_obj(self, dfile, dfile_tiles):
        '''Return gdal object for the GeoImage.'''

        # Need to handle .TIL files specifically because gdal does not fully
        # support them.
        if tt.files.filter(dfile, '*.TIL', case_sensitive=False):
            # If this is a .TIL file, then create a VRT, copy it to an
            # "in memory" obj and then remove the VRT file from disk.

            # Get a temporary file
            file_temp = tempfile.NamedTemporaryFile(suffix=".VRT")

            # Build the vrt command
            cmd = []
            cmd.append("gdalbuildvrt")
            cmd.append(file_temp.name)
            for i in dfile_tiles: cmd.append(i)

            # Execute the buildvrt command
            tt.cmd_line.exec_cmd(cmd)
            # gdal does not issue errors to the command line,
            # so try/except on tt.cmd_line.exec_cmd won't work...
            # Check that file exists to see if the buildvrt succeeded.
            if not os.path.isfile(file_temp.name):
                raise StandardError("Creation of file " + file_temp.name + " "
                                    "failed. This could possibly be a "
                                    "write access problem?")

            vvv = gdal.Open(file_temp.name)

            # Create MEM copy of VRT - created "in memory" by passing an empty
            # file name string to the VRT driver.  If I used the "MEM" driver,
            # the full dataset would be read into memory on creation.
            drv = gdal.GetDriverByName('VRT')
            obj = drv.CreateCopy('',vvv)

            # Delete VRT
            file_temp.close()

            if os.path.isfile(file_temp.name):
                raise StandardError("Removal of file " + file_temp.name + " "
                                    "failed. There is something wrong with "
                                    "the .TIL handling.")

        else:
            obj = gdal.Open(self.files.dfile, gdalconst.GA_ReadOnly)

        return obj

    def _set_metadata(self):
        """ Get image metadata."""
        meta_geoimg_dict = read_geo_file_info(self._fobj)

        # Need to handle case of a .TIL that results an in memory VRT.
        # In this case, file_name will be the VRT string when returned
        # from the gdal driver above and does not print well or return the
        # intended information.
        if not os.path.isfile(meta_geoimg_dict['file_name']):
            meta_geoimg_dict['file_name'] = self.files.dfile

        # Add geoio class name to dictionary
        meta_geoimg_dict['class_name'] = self.__class__.__name__

        ### OrderedBunch the metadata from the read_geo_file_info dictionary
        self.meta = tt.bunch.OrderedBunch(meta_geoimg_dict)

        # Set class members
        self.shape = self.meta.shape
        self.resolution = self.meta.resolution

    def __repr__(self):
        """Human readable image summary similar to the R package 'raster'."""
        sss = ''
        su = self.meta

        prefixes = collections.OrderedDict()
        prefixes['Class Name'] = (['class_name'],'')
        prefixes['Driver Name'] = (['driver_name'],'')
        prefixes['Data Type'] = (['gdal_dtype_name'],'')
        prefixes['File Name'] = (['file_name'],'')
        prefixes['File List'] = (['file_list'], '')
        prefixes['Dimensions'] = (['shape'],
                                  ' (nlayers, nrows, ncols)')
        prefixes['Resolution'] = (['resolution'],' (x,y)')
        # The following is inverted because xstart, xend, etc are calculated
        # in pixel space and this is inverted from the North = max, South = min
        # paradigm.
        prefixes['Extent'] = (['extent'],' (xmin, xmax, ymin, ymax)')
        prefixes['Projection String'] = (['pprint_proj_string'],'')
        prefixes['Geo Transform'] = (['geo_transform'],'')
        prefixes['Authority'] = (['authority'], '')

        ### Loop through prefixes and su to print data to screen
        # Gen max length of labels to set prefix length
        prelen = max([len(x) for x in prefixes])
        # Loop through each prefix to put together wrapped string
        for x in prefixes:
            prefix = x+' '*(prelen-len(x))+' : '
            width_set = 80
            wrapper = textwrap.TextWrapper(initial_indent=prefix,
                                           width=width_set,
                                           replace_whitespace=False,
                                           subsequent_indent=' '*len(prefix))

            message = ', '.join([str(su[y]) for y in prefixes[x][0]])
            message = message+prefixes[x][1]

            # Handle different message formats:
            # If message contains new lines, just pass those through to print.
            if message.find('\n') != -1:
                sss = sss + prefix + message.replace('\n','\n'+' '*prelen)+'\n'
            # Else if message is not empty, pass to wrapper.fill
            elif message:
                sss = sss + wrapper.fill(message) + '\n'
            # Else just pass the empty message along with prefix
            else:
                sss = sss + prefix + '\n'

        return sss

    def print_img_summary(self):
        """Echo the object's __repr__ method."""
        print(self.__repr__())

    def __iter__(self):
        '''Yield from default iter_window iterator.'''
        for x in self.iter_window():
            yield x

    def iter_base(self,xoff,yoff,win_xsize,win_ysize,**kwargs):
        '''
        Base iterator function to yield data from array-like window parameters.

        Parameters
        ----------
        xoff : array_like
            x offset(s) for the image regions to be read.
        yoff : array_like
            y offset(s) for the image regions to be read.
        win_xsize : array_like
            window x-dim size(s) for the image regions to be read.
        win_ysize : array_like
            window y-dim size(s) for the image regions to be read.
        kwargs : optional
            keyword arguments to be passed to get_data.

        Yields
        ------
        ndarray
            Three dimensional numpy array of data from the requested region
            of the image.

        '''

        logger.debug('*** begin iter_base ***')

        # Broadcast array_like
        windows = np.broadcast(xoff,yoff,win_xsize,win_ysize)

        # Iterate through windows generated from input parameters
        for w in windows:
            logger.debug('window parameters: xoff %s, yoff %s, '
                                            'win_xsize %s, win_ysize %s',
                                             w[0], w[1], w[2], w[3])
            yield self.get_data(window=w,**kwargs)

    def iter_window(self, win_size=None, stride=None,**kwargs):
        '''
        Window iterator that yields data from the image based on win_size
        and stride.

        win_size and stride are both optional arguments.  If neither are
        passed, then the method pulls win_size from GDAL GetBlockSize() and
        uses that to step through the image.  If only win_size is provided,
        the method yields adjoining windows of the size requested.  If only
        stride is provided, an error is rasied.

        Parameters
        ----------
        win_size : array-like, length 2, optional
            The size of the requested image chip in x and y.
        stride : array-like, length 2, optional
            The size of the step between each yielded chip in x and y.
        kwargs: optional
            Arguments for get_data().

        Yields
        ------
        ndarray
            Three dimensional numpy array of pixel values from the
            requested region of the image.

        '''

        logger.debug('*** begin iter_window ***')

        # Check input values
        if win_size:
            if any(x <= 0 for x in win_size):
                raise ValueError('No value in win_size can be equal '
                                 'to or less than zero.')

        if stride:
            if any(x <= 0 for x in stride):
                raise ValueError('No value in stride can be equal '
                                 'to or less than zero.')

        # if NOT win_size and NOT stride
        # use gdal to figure out block size and then continue on below.
        if not win_size and not stride:
            # Get block size from gdal
            b = self._fobj.GetRasterBand(1)
            win_size = b.GetBlockSize()

        logger.debug('win_size is:  %s, stride is:  %s', win_size, stride)

        # if win_size and NOT stride
        # set stride to make windows adjoining
        if win_size and not stride:
            # Set vars for easy access below
            xs = self.meta.shape[1]
            ys = self.meta.shape[2]
            xsize, ysize = win_size

            # Find starting offsets by identifying the pixels that don't fit in
            # the requested window blocks and then split the different
            # between ends of the image using floor (int) to reduce fractions.
            x_extra_pixels = xs % win_size[0]
            xoff = int(x_extra_pixels / 2.0)
            y_extra_pixels = ys % win_size[1]
            yoff = int(y_extra_pixels / 2.0)

            # Use while True to loop through get_data until outside the image
            xoff_start = xoff
            xsize, ysize = win_size
            while True:
                logger.debug(' xoff is %s,\tyoff is %s', xoff, yoff)
                yield self.get_data(window=[xoff, yoff, xsize, ysize],**kwargs)
                xoff += xsize
                if xoff > self.meta.shape[1]:
                    xoff = xoff_start
                    yoff += ysize
                if yoff > self.meta.shape[2]:
                    break

        # if NOT win_size and stride, raise error
        elif not win_size and stride:
            raise ValueError('Setting stride and not setting win_size is not '
                             'allowed because there is no resonable value to '
                             'set win_size to.  In this case stride can be '
                             'even or odd which could result in alternative '
                             'size return blocks around the center pixel '
                             '(or fractional pixel).')

        # if win_size and stride
        # just do it
        elif win_size and stride:
            # Set vars for easy access below
            xs = self.meta.shape[1]
            ys = self.meta.shape[2]
            xsize, ysize = win_size
            xstride, ystride = stride

            # Find starting offset by identifying pixels that don't fit in
            # the requested size/stride and then split the different between
            # ends of the image using floor (int) to reduce fractions.
            x_extra_pixels = (xs - xsize) % xstride
            xoff = int(x_extra_pixels/2.0)
            y_extra_pixels = (ys - ysize) % ystride
            yoff = int(y_extra_pixels/2.0)

            # Start the yield loop
            xoff_start = xoff
            while True:
                logger.debug(' xoff is %s,\tyoff is %s', xoff, yoff)
                yield self.get_data(window=[xoff, yoff, xsize, ysize], **kwargs)
                xoff += xstride
                if xoff > self.meta.shape[1]:
                    xoff = xoff_start
                    yoff += ystride
                if yoff > self.meta.shape[2]:
                    break

    def iter_window_random(self, win_size=None, no_chips=1000, **kwargs):
        """Random chip iterator.

        Parameters
        ----------
        win_size : array-like, length 2
            The size of the requested image chip in x and y.
        no_chips : int, optional
            Number of chips to generate.
        kwargs: optional
            Arguments for get_data().

        Yields
        ------
        ndarray
            Three dimensional numpy array of pixel values from the
            requested region of the image.
        """

        # Check input values
        if win_size:
            if any(x <= 0 for x in win_size):
                raise ValueError('No value in win_size can be equal '
                                 'to or less than zero.')

        counter = no_chips
        xs = self.meta.shape[1]
        ys = self.meta.shape[2]
        xsize, ysize = win_size

        while True:
            # select random offset
            xoff = np.random.randint(xs-xsize+1)
            yoff = np.random.randint(ys-ysize+1)
            yield self.get_data(window=[xoff, yoff, xsize, ysize], **kwargs)
            counter -= 1
            if counter == 0: break


    def iter_components(self, **kwargs):
        """This is a convenience method that iterataes (via yield) through
        the components in the image object.  Any kwargs valid for get_data
        can be passed through."""

        for c in xrange(len(self.files.dfile_tiles)):
            yield self.get_data(component=c, **kwargs)


    def iter_vector(self, vector=None, properties=False, filter=None, **kwargs):
        """This method iterates (via yeild) through a vector object or file.
        Any kwargs valid for get_data can be passed through."""

        if 'window' in kwargs.keys():
            raise ValueError("The window argument is not valid for this " \
                             "method. They both define a retrieval " \
                             "geometry.  Pass one or the other.")

        if 'geom' in kwargs.keys():
            raise ValueError("The geom argument is not valid for this " \
                             "method. The vector file passed in defines " \
                             "the retrieval geometry.")

        # ToDo Test for overlap of geom and image data?

        obj = ogr.Open(vector)
        lyr = obj.GetLayer(0)
        lyr_sr = lyr.GetSpatialRef()

        img_proj = self.meta.projection_string
        img_trans = self.meta.geo_transform
        img_sr = osr.SpatialReference()
        img_sr.ImportFromWkt(img_proj)

        coord_trans = osr.CoordinateTransformation(lyr_sr, img_sr)

        for feat in lyr:
            # Return feature properties data is requested
            if properties is True:
                prop_out = feat.items()
            elif properties:
                if isinstance(properties, (list, tuple, str)):
                    if isinstance(properties, str):
                        properties = [properties]
                    it = feat.items()
                    if not all(x for x in properties if x in it.keys()):
                        raise ValueError("One or more of the requested "
                                         "properties are not in the vector "
                                         "feature.")
                    prop_out = {x: it[x] for x in properties if x in it.keys()}
                    if not prop_out:
                        prop_out = None
                        warnings.warn("No properties value found matching "
                                      "request.")
                else:
                    raise ValueError("Invalid properties argument.")


            # Determine if the feature should be returned based on value of
            # filter and if the value exists in the feature properties.
            if filter:
                # The filter should be either a list of dictionary key/value
                # paris of length one, or of list of key/value pairs.  The
                # idea is that you can filter against more than one value
                # of a key when you can pass a list of pairs.
                if isinstance(filter, dict) & (len(filter) != 1):
                    raise ValueError("Filters should be passed in as a " \
                                     "list of dictionaries that will " \
                                     "be used to filter against the " \
                                     "feature property values.")

                # If filter is a dict of len 1, convert to a list of len 1 for
                # the looping code below.
                if isinstance(filter,dict):
                    filter = [filter]

                # Get feature properties to check against.
                prop_test = feat.items()

                # raise warning if a filter item key is not in properties
                if any(f.keys()[0] not in prop_test.keys() for f in filter):
                    warnings.warn("Requested filter key is not present in "
                                  "vector properties.")

                # If any filter is caught, pass on, otherwise return
                if any(prop_test.get(d.keys()[0], None) ==
                                    d.values()[0] for d in filter):
                    pass
                else:
                    if properties:
                        yield (None, None)
                        continue
                    else:
                        yield None
                        continue

            # Get the geometry to pass to get_data
            geom = feat.geometry()

            # Use transform from above to put geom in image space
            geom.Transform(coord_trans)

            # Catch and pass OverlapError for the iterator
            try:
                data = self.get_data(geom=geom, **kwargs)
            except OverlapError:
                data = None

            # Yield the data
            if properties:
                yield (data, prop_out)
            else:
                yield data

    def get_data_from_vec_extent(self, vector=None, **kwargs):
        """This is a convenience method to find the extent of a vector and
        return the data from that extent.  kwargs can be anything accepted
        by get_data."""
        if vector is None:
            raise ValueError("Requires a vector to read.  The vector can be " \
                             "a string that describes a vector object or a " \
                             "path to a valid vector file.")

        if 'window' in kwargs.keys():
            raise ValueError("The window argument is not valid for this " \
                             "method. The vector file passed in defines " \
                             "the retrieval geometry.")

        if 'geom' in kwargs.keys():
            raise ValueError("The geom argument is not valid for this " \
                             "method. The vector file passed in defines " \
                             "the retrieval geometry.")

        if 'mask' in kwargs.keys():
            raise ValueError("A mask request is not valid for this method " \
                             "because it retrives data from the full extent " \
                             "of the vector.  You might want a rasterize " \
                             "method or iter_vector?")

        # ToDo Test for overlap of geom and image data?

        obj = ogr.Open(vector)
        lyr = obj.GetLayer(0)
        lyr_sr = lyr.GetSpatialRef()

        img_proj = self.meta.projection_string
        img_trans = self.meta.geo_transform
        img_sr = osr.SpatialReference()
        img_sr.ImportFromWkt(img_proj)

        coord_trans = osr.CoordinateTransformation(lyr_sr,img_sr)

        extent = lyr.GetExtent()

        window = self._extent_to_window(extent,coord_trans)
        [xoff, yoff, win_xsize, win_ysize] = window

        return self.get_data(window = window, **kwargs)

    def _extent_to_window(self,extent,coord_trans=None):

        if not coord_trans:
            warnings.warn('No projection checking is done.  Returning passed '
                          'geometry in image space.')

        [minX, maxX, minY, maxY] = extent
        ul_vec = [minX, maxY]
        lr_vec = [maxX, minY]

        if coord_trans:
            # Online dobumentation says that there could be a bug in
            # TransformPoints - seems to be working with 1.11.4 so
            # I'll leave it in for now.  When working with a bag vector file
            # TransformPoints returned inf values that exploded below whereas
            # the sequental TransformPoint returned values that triggered
            # the overlap error below.  Leaving the single call for now
            # as it is faster and cleaner.
            [ul_img, lr_img] = coord_trans.TransformPoints([ul_vec, lr_vec])
            #ul_img = coord_trans.TransformPoint(*ul_vec)
            #lr_img = coord_trans.TransformPoint(*lr_vec)

            # Cut off third returned dimension (should be zero)
            ul_img = ul_img[:-1]
            lr_img = lr_img[:-1]
        else:
            ul_img = ul_vec
            lr_img = lr_vec

        xs,ys = self.proj_to_raster(*zip(*[ul_img,lr_img]))

        xoff = int(math.floor(min(xs)))
        yoff = int(math.floor(min(ys)))

        xmax = int(math.ceil(max(xs)))
        ymax = int(math.ceil(max(ys)))

        win_xsize = xmax-xoff
        win_ysize = ymax-yoff

        window = [xoff, yoff, win_xsize, win_ysize]

        # Logging info if needed
        logger.debug('vector geo extent...\n\t%s\n\t%s',ul_vec,lr_vec)
        logger.debug('image geo extent...\n\t%s\n\t%s',ul_img,lr_img)
        logger.debug('geo transform...\n\t%s', self.meta.geo_transform)
        logger.debug('raster xy extent...\n\t%s,\n\t%s',xs,ys)
        logger.debug('requested window...\n\t%s',window)

        if ((xoff + win_xsize <= 0) or (yoff + win_ysize <= 0) or
            (xoff > self.meta.shape[1]) or (yoff > self.meta.shape[2])):
            raise OverlapError("The requested data window has no " \
                              "content.  Perhaps the image and vector " \
                              "do not overlap or the projections may " \
                              "not be able to be automatically reconciled?")

        return window

    def proj_to_raster(self, projx, projy):
        '''
        Method to convert points in projection space to points in raster space.

        Input can be in a variety of types as long as both the input parameters
        are of the same type.  The method will attempt to return a data type
        as similar as possible to the input type.

        Parameters
        ----------
        projx : int, float, list, tuple, or numpy.ndarray
            Input point in projected space.
        projy : int, float, list, tuple, or numpy.ndarray
            Input point in projected space.

        Returns
        -------
        x : float, list, tuple, or numpy.ndarray
            raster x value calculated from projx, projy, and the object's
            geo_transform
        y : float, list, tuple, or numpy.ndarray
            raster y value calculated from projx, projy, and the object's
            geo_transform
        '''

        # This method can't handle mixed types
        if type(projx) != type(projy):
            raise ValueError('The type of x and y should be the same and '
                             'either integers, float, tuples, lists, or '
                             'numpy arrays.')

        # Determine input type so that the conversion back can be done
        # on return
        intype = 0
        if not hasattr(projx, '__iter__'):
            intype = 1
        elif isinstance(projx, list):
            intype = 2
        elif isinstance(projx, tuple):
            intype = 3
        elif isinstance(projx, np.ndarray):
            intype = 4
        else:
            raise ValueError("The input type was not recognized.")

        # Convert to numpy arrays for the calculation
        projx = np.asarray(projx)
        projy = np.asarray(projy)

        # Get geo_transform from object
        gm = self.meta.geo_transform

        # Transform per inverse of http://www.gdal.org/gdal_datamodel.html
        x = (gm[5] * (projx - gm[0]) - gm[2] * (projy - gm[3])) / \
            (gm[5] * gm[1] + gm[4] * gm[2])
        y = (projy - gm[3] - x * gm[4]) / gm[5]

        # Return to input type
        if intype == 1:
            return float(x), float(y)
        elif intype == 2:
            return list(x), list(y)
        elif intype == 3:
            return tuple(x), tuple(y)
        elif intype == 4:
            return x, y

    def raster_to_proj(self, x, y):
        '''
        Method to convert points in raster space to points in projection space.

        Input can be in a variety of types as long as both the input parameters
        are of the same type.  The method will attempt to return a data type
        as similar as possible to the input type.

        Parameters
        ----------
        x : int, float, list, tuple, or numpy.ndarray
            Input point in raster space.
        y : int, float, list, tuple, or numpy.ndarray
            Input point in raster space.

        Returns
        -------
        projx : float, list, tuple, or numpy.ndarray
            projection x value calculated from x, y, and the object's
            geo_transform
        projy : float, list, tuple, or numpy.ndarray
            projection y value calculated from x, y, and the object's
            geo_transform
        '''

        # This method can't handle mixed types
        if type(x) != type(y):
            raise ValueError('The type of x and y should be the same and '
                             'either integers, float, tuples, lists, or '
                             'numpy arrays.')

        # Determine input type so that the conversion back can be done
        # on return
        intype = 0
        if not hasattr(x, '__iter__'):
            intype = 1
        elif isinstance(x, list):
            intype = 2
        elif isinstance(x, tuple):
            intype = 3
        elif isinstance(x, np.ndarray):
            intype = 4
        else:
            raise ValueError("The input type was not recognized.")

        # Convert to numpy arrays for the calculation
        x = np.asarray(x)
        y = np.asarray(y)

        # Get geo_transform from object
        gm = self.meta.geo_transform

        # Transform per http://www.gdal.org/gdal_datamodel.html
        projx = gm[0] + gm[1] * x + gm[2] * y
        projy = gm[3] + gm[4] * x + gm[5] * y

        # Return to input type
        if intype == 1:
            return float(projx), float(projy)
        elif intype == 2:
            return list(projx), list(projy)
        elif intype == 3:
            return tuple(projx), tuple(projy)
        elif intype == 4:
            return projx, projy

    def get_data(self, component = None,
                       bands = None,
                       window = None,
                       buffer = None,
                       geom = None,
                       mask = False,
                       mask_all_touched = False,
                       virtual = False):
        """Read data from geo-image file.  If component is specified and
        this is a .vrt or .til file, then it will pull only the data from
        the file specified in self.dfile_tiles.  Component is specified base 1.

        The isn't generally a reason to call both component and window,
        but it isn't explicitly disallowed.  If window and component are both
        specified, the resulting data window is relative to the coordinates
        of the specified component."""

        if component is not None:
            if component == 0:
                raise ValueError("Component should be specified as based 1.")

            if component > len(self.files.dfile_tiles):
                raise ValueError("You've requested a component value greater "
                                 "than the number available.")

            y = GeoImage(self.files.dfile_tiles[component-1])
            logger.debug('returning data from:  '+
                  str(self.files.dfile_tiles[component-1]))
            obj = y._fobj
        else:
            obj = self._fobj

        # If bands not requested, set to pull all bands
        if not bands:
            bands = range(1,obj.RasterCount+1)

        # Set window to pull
        if window and geom:
            raise ValueError("The arguments window and geom are mutually " \
                             "exclusive.  They both define an image " \
                             "extent.  Pass either one or the other.")

        if window:
            # Set extent paramets based on window if provided
            if len(window) == 4:
                [xoff, yoff, win_xsize, win_ysize] = window
            else:
                raise ValueError("Window must be length four and will be read" \
                                 "as: xoff, yoff, win_xsize, win_ysize")
        elif geom:
            # Set window size based on a geom object in image space
            # ToDo - Add all_touched option to this and mask function.
            g = self._instantiate_geom(geom)
            extent = g.GetEnvelope()
            window = self._extent_to_window(extent)
            [xoff, yoff, win_xsize, win_ysize] = window
        else:
            # Else use extent of image to set extent params
            xoff = 0
            yoff = 0
            win_xsize = self.meta.shape[1]
            win_ysize = self.meta.shape[2]

        # Add buffer
        if buffer:
            if isinstance(buffer,int):
                buffer = [buffer]

            if len(buffer) == 1:
                xbuff = buffer[0]
                ybuff = buffer[0]
            elif len(buffer) == 2:
                xbuff = buffer[0]
                ybuff = buffer[1]
            else:
                raise ValueError("Buffer must be either length one or two.")

            # Apply the buffer to the readasarray parameters
            xoff = xoff-xbuff
            yoff = yoff-ybuff
            win_xsize = win_xsize+2*xbuff
            win_ysize = win_ysize+2*ybuff

        # Handle out-of-bounds cases
        # (i.e. xoff = 0; buffer = 3; xoff - buffer)
        # initialize buffer vars
        np_xoff_buff = 0
        np_yoff_buff = 0
        np_xlim_buff = 0
        np_ylim_buff = 0

        if xoff < 0:
            np_xoff_buff = xoff
            win_xsize = win_xsize + xoff
            xoff = 0

        if yoff < 0 :
            np_yoff_buff = yoff
            win_ysize = win_ysize + yoff
            yoff = 0

        xpos = xoff+win_xsize
        xlim = self.meta.shape[1]
        ypos = yoff+win_ysize
        ylim = self.meta.shape[2]

        if xpos > xlim:
            np_xlim_buff = xpos-xlim
            win_xsize = win_xsize-np_xlim_buff
        if ypos > self.meta.shape[2]:
            np_ylim_buff = ypos-ylim
            win_ysize = win_ysize-np_ylim_buff

        # # This code will just buffer window requests outside of the
        # # image dimension, so I need to explicitly catch bad requests
        # if np.abs(np_xoff_buff) > xbuff or \
        #    np.abs(np_xlim_buff) > xbuff or \
        #    np.abs(np_yoff_buff) > ybuff or \
        #    np.abs(np_ylim_buff) > ybuff:
        #    raise ValueError("Requested window is outside the image.")

        # Read data
        if virtual is True:
            raise NotImplementedError('keyword argument not implemented yet.')
        elif virtual is False:
            # Read data one band at a time with ReadAsArray
            for i,b in enumerate(bands):
                bobj = obj.GetRasterBand(b)
                # The try/except below is used to initialize the data variable...
                # The first loop trigger the NameError, so the except is used to
                # then initialize "data".
                try:
                    data[i,:,:] = bobj.ReadAsArray(xoff=xoff,
                                                   yoff=yoff,
                                                   win_xsize=win_xsize,
                                                   win_ysize=win_ysize)
                except NameError:
                    zt = len(bands)
                    dt = const.DICT_GDAL_TO_NP[bobj.DataType]
                    data = np.empty([zt, win_ysize, win_xsize], dtype=dt)
                    data[i,:,:] = bobj.ReadAsArray(xoff=xoff,
                                                   yoff=yoff,
                                                   win_xsize=win_xsize,
                                                   win_ysize=win_ysize)
        else:
            raise ValueError("virtual keyword argument should be boolean.")

        # Convert numpy array to masked numpy array if requested.
        if mask and geom:
            # Set image parameters
            xres = self.meta.resolution[0]
            yres = self.meta.resolution[1]
            (xmin, xmax, ymin, ymax) = g.GetEnvelope()
            ul_env = [xmin, ymax]
            ul_raster = self.proj_to_raster(*ul_env)
            ul_corner = [math.floor(ul_raster[0]),math.floor(ul_raster[1])]
            xmin_corner,ymax_corner = self.raster_to_proj(*ul_corner)

            # Create temporary raster to burn
            drv = gdal.GetDriverByName('MEM')
            tds = drv.Create('', win_xsize, win_ysize, 1, gdal.GDT_Byte)
            tds.SetGeoTransform((xmin_corner, xres, 0, ymax_corner, 0, -yres))
            tds.SetProjection(self.meta.projection_string)

            # Create ogr layr from geom
            odrv = ogr.GetDriverByName('Memory')
            ds = odrv.CreateDataSource('')

            ltype = g.GetGeometryType()
            lsrs = osr.SpatialReference(self.meta.projection_string)
            lyr = ds.CreateLayer('burnshp', lsrs, ltype)

            feat = ogr.Feature(lyr.GetLayerDefn())
            feat.SetGeometryDirectly(g)
            lyr.CreateFeature(feat)

            # Run the burn
            if mask_all_touched:
                err = gdal.RasterizeLayer(tds, [1], lyr, burn_values=[1],
                                                options=['ALL_TOUCHED=TRUE'])
            else:
                err = gdal.RasterizeLayer(tds, [1], lyr, burn_values=[1])

            # build the masked array
            m = tds.ReadAsArray().astype('bool')
            data = np.ma.array(data, mask=np.tile(~m, (data.shape[0], 1, 1)))

        if mask and not geom:
            # This code will mask values outside the image as well as zeros
            # inside the image.
            data = np.ma.array(data,mask=~data.astype('bool'))
            # The line below will only mask values outside the image.
            # data = np.ma.array(data,mask=np.zeros(data.shape).astype('bool'))

        # Pad the output array if needed
        pad_tuples = ((0,0),
                      (np.abs(np_yoff_buff), np.abs(np_ylim_buff)),
                      (np.abs(np_xoff_buff), np.abs(np_xlim_buff)))
        if not mask:
            data = np.pad(data, pad_tuples, 'constant',constant_values=0)
        elif mask:
            mpad = np.pad(data.mask, pad_tuples, 'constant', constant_values=1)
            data = np.pad(data, pad_tuples, 'constant', constant_values=0)
            data = np.ma.array(data,mask=mpad)

        # if "y" is open, close it
        try:
            y = None
        except NameError:
            pass

        return data

    def _instantiate_geom(self,g):
        """Attempt to convert the geometry pass in to and ogr Geometry
        object.  Currently implements the base ogr.CreateGeometryFrom*
        methods and will reform fiona geometry dictionaries into a format
        that ogr.CreateGeometryFromJson will correctly handle.
        """

        if isinstance(g,ogr.Geometry):
            # If the input geometry is already an ogr object, create a copy
            # of it.  This is requred because of a subtle issue that causes
            # gdal to crash if the input geom is used elsewhere.  The working
            # theory is that the geometry is freed when going out of a scope
            # while it is needed in the upper level loop.  In this code, the
            # problem comes about between self.iter_vector and self.get_data
            # with mask=True.
            return ogr.CreateGeometryFromJson(str(g.ExportToJson()))

        # Handle straight ogr GML
        try:
            return ogr.CreateGeometryFromGML(g)
        except:
            pass
        # Handle straight ogr Wkb
        try:
            return ogr.CreateGeometryFromWkb(g)
        except:
            pass
        # Handle straight ogr Json
        try:
            return ogr.CreateGeometryFromJson(g)
        except:
            pass
        # Handle straight ogr Wkt
        try:
            return ogr.CreateGeometryFromWkt(g)
        except:
            pass
        # Handle fiona Json geometry format
        try:
            gjson = str(g).replace('(','[').replace(')',']')
            return ogr.CreateGeometryFromJson(gjson)
        except:
            pass

        raise ValueError("A geometry object was not able to be created from " \
                         "the value passed in.")

    def write_img_like_this(self,new_fname,np_array,return_obj=False,
                            gdal_driver_name=None,options=[],
                            vrt_fallback="GTiff"):
        """Write the data passed in the variable "np_array" as a new image
        with image parameters (projection, etc.) pulled from this object.
        The new file is created as the data type of the numpy array
        "np_array".  "np_array" should have the same line/sample size as
        the object data, but the number of bands can vary.  This method uses
        the create_geo_image function in this module.  That function uses
        gdal create and has minimal metadata copy since what should or
        should not be copied is application dependent.
        """

        # If the data array is 2D, add a dimension for a single band
        if np_array.ndim == 2:
            np_array = np_array[np.newaxis, :, :]
        elif np_array.ndim == 3:
            pass
        else:
            raise ValueError("This can't handle Arrays larger than three "
                             "dimensions.")

        if not ((np_array.shape[1] == self.shape[2]) and \
                (np_array.shape[2] == self.shape[1])):
            raise ValueError("Data in is not the same size as that in the "
                             "current image object.  Data in is shape: %s, "
                             "Object shape is: %s" %
                             (np_array.shape, self.shape))

        if gdal_driver_name is None:
            gdal_driver_name = self.meta.driver_name

        ## Write the geo image using my geoio function
        # NDV set to 0 by default in create_geo_image
        new_img_name = create_geo_image(new_file_name = new_fname,
                                        data_np_array = np_array,
                                        gdal_driver_name = gdal_driver_name,
                                        #gdal_driver_name = 'GTiff',
                         gdal_geo_t = self.meta.geo_transform,
                                        gdal_projection = self.meta.projection_string,
                                        data_type = np_array.dtype,
                                        NDV=self.meta.no_data_value,
                                        options=options,
                                        vrt_fallback=vrt_fallback)

        # Output information about the new file
        f = GeoImage(new_img_name)

        logger.debug('')
        logger.debug("######################")
        logger.debug("New image file has been created at:  "+new_img_name)
        logger.debug("Data type is:  "+str(np_array.dtype))
        if self.meta.no_data_value:
            logger.debug("No data value set to:  " +
                         str(self.meta.no_data_value))
        else:
            logger.debug("No data value was not set.")
        logger.debug("######################")

        if return_obj:
            return f

    def write_img_replace_this(self,np_array):
        """Replace the data in the current object image with the data passed
        in the variable "np_array".  This method uses gdal to replace the
        data in place - so all metadata is intact.
        """
        # All dims must be exact
        if not ((np_array.shape[0] == self.shape[0]) and \
                (np_array.shape[1] == self.shape[2]) and \
                (np_array.shape[2] == self.shape[1])):
            raise ValueError("Data in is not the same shape as that in the "
                             "current image object.  Data in is shape: %s, "
                             "Object shape is: %s" %
                             (np_array.shape, self.shape))

        # Check that the incoming array is the same dtype as the image.
        gi_dtype = self.meta.gdal_dtype
        if not (np_array.dtype == const.DICT_GDAL_TO_NP[gi_dtype]):
            raise ValueError("Data types must match exactly to do a "
                             "data replace.")

        # Reload gdal object in update mode
        reload_fname = self.meta_fname
        self._fobj = None
        self._fobj = gdal.Open(reload_fname, gdalconst.GA_Update)

        # Load new data into gdal object file
        nbands = self.shape[0]
        for i in xrange(nbands):
            b = self._fobj.GetRasterBand(i+1).WriteArray(np_array[i,:,:])
            b = None

        # Dereference gdal object and reopen as read only
        self._fobj = None
        self._fobj = gdal.Open(reload_fname, gdalconst.GA_ReadOnly)

    # ToDo Add ability to repoject data during file write.

    def get_stretch_values(self,**kwargs):
        # Call the stretch_values function in this module.
        return get_img_stretch_vals(self._fobj,**kwargs)

def read_geo_file_info(fname_or_fobj):
    """ Get image metadata."""
    # class       : RasterBrick
    # dimensions  : 3191, 921, 2938911, 11  (nrow, ncol, ncell, nlayers)
    # resolution  : 30, 30  (x, y)
    # extent      : 630885, 658515, 2796285, 2892015  (xmin, xmax, ymin, ymax)
    # coord. ref. : +proj=utm +zone=40 +datum=WGS84 +units=m +no_defs +ellps=WGS84 +towgs84=0,0,0
    # data source : /path/to/file/desertEO1H1580422003134110PZ.TIFF
    # names       : desertEO1H1580422003134110PZ.1, desertEO1H1580422003134110PZ.2, desertEO1H1580422003134110PZ.3, desertEO1H1580422003134110PZ.4, desertEO1H1580422003134110PZ.5, desertEO1H1580422003134110PZ.6, desertEO1H1580422003134110PZ.7, desertEO1H1580422003134110PZ.8, desertEO1H1580422003134110PZ.9, desertEO1H1580422003134110PZ.10, desertEO1H1580422003134110PZ.11

    # If the filename passed in is a string, try to open the file, else you
    # can assume that it is a file object already opened and passed in.
    if isinstance(fname_or_fobj,str):
        fobj = gdal.Open(fname_or_fobj, gdalconst.GA_ReadOnly)
    else:
        fobj = fname_or_fobj

    summary = {}
    summary['file_name'] = fobj.GetDescription()
    summary['file_list'] = fobj.GetFileList()
    summary['driver_name'] = fobj.GetDriver().ShortName
    summary['no_data_value'] = fobj.GetRasterBand(1).GetNoDataValue()
    summary['gdal_dtype'] = fobj.GetRasterBand(1).DataType
    summary['gdal_dtype_name'] = gdal.GetDataTypeName(summary['gdal_dtype'])

    ### Get Image basics dimensions
    x = fobj.RasterXSize
    y = fobj.RasterYSize
    summary['pixels'] = x*y
    nbands = fobj.RasterCount
    summary['shape'] = (nbands, x, y)

    ### Get image resolution and extents
    # Pulled from GDALInfoReportCorner() at:
    # http://www.gdal.org/gdalinfo_8c.html
    gt = fobj.GetGeoTransform()
    summary['geo_transform'] = gt
    xstart = gt[0]
    xend = gt[0] + gt[1]*x + gt[2]*y
    ystart = gt[3]
    yend = gt[3] + gt[4]*x + gt[5]*y

    # Pixel Resolutions
    xres = abs(gt[1])
    yres = abs(gt[5])

    # Add to summary
    summary['resolution'] = (xres, yres)
    summary['extent'] = (xstart, xend, ystart, yend)

    ### Get image projection and datum
    summary['projection_string'] = fobj.GetProjection()
    # Get pretty printable projection_string
    sr = osr.SpatialReference(fobj.GetProjectionRef())
    summary['pprint_proj_string'] = sr.ExportToPrettyWkt()
    summary['authority'] = sr.GetAttrValue("AUTHORITY",0)+ ':' + \
                           sr.GetAttrValue("AUTHORITY",1)

    return summary

# Function to write a new file.
def create_geo_image(new_file_name, data_np_array, gdal_driver_name,
                     gdal_geo_t, gdal_projection, data_type, NDV=0,
                     options=[],vrt_fallback="GTiff"):
    """ Create a new image file with all the parameters pass in.  This
    function is implemented in the GeoImage Class.  It will autopopulate the
    parameters so is a bit easier to use from there.
    """

    if gdal_driver_name == "VRT":
        # Change driver name to something that will write to disk.  This
        # defaults to GTiff but is configurable.
        logger.debug("You have requested a VRT.  Creating a single new real "
                     "file using %s format.  This format can be ovridden "
                     "in the call using 'GTiff', 'ENVI', etc.  Eventually, "
                     "this could be replaced with a file-for-file VRT "
                     "rewrite, but that does not always make sense in VRTs "
                     "since there can be pixels in the original image "
                     "(overlaps) that aren't in the "
                     "new data array.",vrt_fallback)
        gdal_driver_name = vrt_fallback

        # Strip ".vrt" extension if it exists and rename according to the
        # driver that has been substituted.
        # --- Only supports ENVI and GTiff now, I can't figure out how to do
        #  this dynamically in gdal.  (as of 141218)
        if tt.files.filter(new_file_name, '*.VRT', case_sensitive=False):
            if gdal_driver_name == "GTiff":
                new_file_name = os.path.splitext(new_file_name)[0] + ".TIF"
            elif gdal_driver_name == "ENVI":
                new_file_name = os.path.splitext(new_file_name)[0]
            else:
                new_file_name = os.path.splitext(new_file_name)[0]

    ## Convert data types to the appropriate value.  The gdal create
    ## routine can pass either the gdal data type int or the alias (such as
    ## gdal.GDT_Float32).
    # If the data type passed is of type numpy, convert it to GdalDataType
    if data_type in const.DICT_NP_TO_GDAL:
        data_type = const.DICT_NP_TO_GDAL[data_type]
    # If gdal dtype class then pass
    elif data_type in const.DICT_GDAL_TO_NP:
        # Maybe convert to the integer form?
        pass
    # if gdal dtype name, then convert to data type integer alias
    elif isinstance(data_type,str):
        data_type = gdal.GetDataTypeByName(data_type)

    # If the data array is 2D, add a dimension for a single band
    if data_np_array.ndim == 2:
        data_np_array = data_np_array[np.newaxis, :, :]
    elif data_np_array.ndim == 3:
        pass
    else:
        raise ValueError("This can't handle Arrays larger than three "
                         "dimensions.")
    (n_bands,y_size,x_size) = data_np_array.shape

    # Create driver and data set object
    driver = gdal.GetDriverByName(gdal_driver_name)
    # Check that the driver supports the data_np_array.dtype
    dstr = gdal.GetDataTypeName(const.DICT_NP_TO_GDAL[data_np_array.dtype])
    dlist = driver.GetMetadata()['DMD_CREATIONDATATYPES'].split()
    if not dstr in dlist:
        raise ValueError("The data type of the provided numpy array is not "
                         "supported by the requested file format.")
    if options:
        if not isinstance(options,list):
            raise ValueErrror("The options list is malformed.  It should be a "
                              "list of strings")
    dst_ds = driver.Create(new_file_name, x_size, y_size, n_bands,
                           data_type, options)

    # Set projection and geotransform
    dst_ds.SetGeoTransform(gdal_geo_t)
    dst_ds.SetProjection(gdal_projection)

    ### Write the new data
    # Set nans to the original No Data Value
    if NDV is not None:
        data_np_array[np.isnan(data_np_array)] = NDV
        data_np_array[np.isinf(data_np_array)] = NDV
        ### Other???

    # Write the bands
    for b in range(n_bands):
        dst_ds.GetRasterBand(b+1).WriteArray( data_np_array[b,:,:] )
        if NDV is not None:
            dst_ds.GetRasterBand(b+1).SetNoDataValue( NDV )

    # Once we're done, close properly the dataset
    dst_ds = None

    # Return the new file name in case it was changed due to vrt fallback.
    return new_file_name

def get_img_stretch_vals(imgfname_or_gdalobj,stretch=[0.02,0.98],approx_ok=True):
    """Read image, mask very small values, and return max and min

    imgfname_or_gdalobj: Either a file name string of a gdal object
    stretch:               stretch range to return in a list
    approx_ok:           Is it ok to approximate the max/min values (bool)
    hist_nbins:          How many bins the histogram is made of to find the
                         appropriate cut value (more bins is more accurate,
                         but slower).
    """

    # ToDo Add tests for get_img_strech_vals
    ### Cases to consider:
    # Lots of fill 0 values with much larger data
    # Lots of fill 0 values with data near zero
    # Large integers
    # Floating point 0-1
    # Floating point larger than 1
    # Negative values to postivie values (a la NDVI)
    #... basically this needs to respect the No Data Value and let the math fly

    if isinstance(imgfname_or_gdalobj,str):
        f = gdal.Open(imgfname_or_gdalobj)
    elif isinstance(imgfname_or_gdalobj,gdal.Dataset):
        f = imgfname_or_gdalobj
    else:
        raise ValueError("The variable passed in as imgfname_or_gdalobj is "
                         "neither an image filename string or a gdal object.")

    # Setup for query of each band
    outofrange = 0  # value zero means no, don't include out of range
    hist_nbins = 1000
    if approx_ok:
        approx = 1
    else:
        approx = 0

    for i in range(1,f.RasterCount+1):
        # ToDo Add respect for NoDataValue

        # Get gdal band object
        b = f.GetRasterBand(i)

        # Get band max/min
        tmp = b.ComputeRasterMinMax(approx)
        imgmax = tmp[1]
        imgmin = tmp[0]

        # Use max/min to get histogram
        hist = b.GetHistogram(imgmin,imgmax,hist_nbins,outofrange,approx_ok)
        hist = np.asarray(hist)
        bins = np.linspace(imgmin, imgmax, len(hist))

        # Get stretch values
        cdf = hist.cumsum()
        top = cdf.max() * stretch[1]
        bottom = cdf.max() * stretch[0]

        topcut = bins[np.where(cdf > top)[0][0]]
        bottomcut = bins[np.where(cdf < bottom)[0][-1]]

    return (bottomcut,topcut)




import unittest
import os
import collections
from osgeo import gdalconst
import numpy as np
import warnings

import geoio.dg
import tinytools as tt
import geoio
from geoio import constants as const
import dgsamples

class TestGeoioEnv(unittest.TestCase):
    """Testing for :
    geoio environment
    """

    def test_for_gdal(self):
        #tt.cmd_line.exec_cmd(['gdalinfo','--version'])
        # This is basically testing the the function runs without errors
        #self.assertTrue(True)
        self.assertTrue(tt.cmd_line.is_tool('gdalinfo'))

    def test_for_const(self):
        self.assertEqual(const.DG_ABSCAL_GAIN['WV02_P'], [0.960])
        self.assertEqual(const.DG_ABSCAL_GAIN['WV03_MULTI'][2], 0.907)

    def test_const_dict_np_to_gdal(self):
        a = np.array([1.23],dtype='float32')
        self.assertEqual(const.DICT_NP_TO_GDAL[a.dtype], gdalconst.GDT_Float32)

    def test_const_dict_gdal_to_np(self):
        a = np.array([1.23],dtype='float32')
        self.assertEqual(const.DICT_GDAL_TO_NP[gdalconst.GDT_Float32], a.dtype)

class TestGeoImage(unittest.TestCase):
    """Testing for :
    GeoImage class
    """
    # Run defined method setUp to setup the test environment
    def setUp(self):
        self.test_img = dgsamples.wv2_longmont_1k.ms
        self.img = geoio.GeoImage(self.test_img)

    def tearDown(self):
        # Remove gdal image object
        self.img = None
        # Find temp files from test directory and remove
        tmp_files=tt.files.search(os.path.dirname(os.path.realpath(__file__)),
                        ["*temp*","*tmp*"],case_sensitive=False)

        print(tmp_files)
        [os.remove(f) for f in tmp_files]

    def test_GeoImage_files_meta_exists(self):
        self.assertIsInstance(self.img.files,tt.bunch.OrderedBunch)

    def test_GeoImage_meta_exists(self):
        self.assertIsInstance(self.img.meta,tt.bunch.OrderedBunch)

    def test_GeoImage_shape(self):
        self.assertEqual(self.img.shape,(8,500,501))

    def test_GeoImage_resolution(self):
        self.assertEqual(self.img.resolution[0],2)
        self.assertEqual(self.img.resolution[1],2)

    def test_GeoImage_get_data(self):
        a = self.img.get_data()
        self.assertTrue(np.array_equal(a[:,100,100],[472,379,563,690,364,696,447,734]))
        self.assertTrue(np.array_equal(a[:,200,200],[575,439,659,784,402,742,449,706]))
        self.assertTrue(np.array_equal(a[:,300,300],[313,192,253,279,161,376,308,571]))
        self.assertTrue(np.array_equal(a[:,400,400],[322,199,258,283,143,356,287,495]))

    def test_GeoImage_get_data_dtypes(self):
        a = self.img.get_data()
        self.assertEqual(const.DICT_NP_TO_GDAL[a.dtype],
                         self.img.meta.gdal_dtype)

    def test_GeoImage_write_img_like_this(self):
        a = (self.img.get_data()*0.01).astype('float32')
        back = self.img.write_img_like_this("tmp.tif",a,return_obj=True)
        b = back.get_data()
        self.assertIsNone(np.testing.assert_array_almost_equal(a,b,decimal=6))

class TestGeoImage_getdata(unittest.TestCase):
    # Run defined method setUp to setup the test environment
    def setUp(self):
        self.test_img = dgsamples.wv2_longmont_1k.ms
        self.img = geoio.GeoImage(self.test_img)

    def tearDown(self):
        # Remove gdal image object
        self.img = None
        # Find temp files from test directory and remove
        tmp_files = tt.files.search(os.path.dirname(os.path.realpath(__file__)),
                                    ["*temp*", "*tmp*"], case_sensitive=False)

        print(tmp_files)
        [os.remove(f) for f in tmp_files]

    def test_get_data_no_options(self):
        d = self.img.get_data()
        self.assertEqual(d.shape,(8,501,500))

    def test_get_data_bands_num(self):
        d = self.img.get_data(bands=[7,3,1])
        self.assertEqual(d.shape,(3,501,500))

    def test_get_data_window_centered(self):
        d = self.img.get_data(window=[10,10,21,23])
        self.assertEqual(d.shape,(8,23,21))

    def test_get_data_window_edge(self):
        d = self.img.get_data(window=[499,500,1,1])
        known = np.array(
            [[[354]],[[226]],[[304]],[[341]],[[176]],[[430]],[[312]],[[518]]])
        self.assertTrue(np.array_equal(d,known))

class TestGeoImage_iter_vector(unittest.TestCase):
    # Run defined method setUp to setup the test environment
    def setUp(self):
        self.test_img = dgsamples.wv2_longmont_1k.ms
        self.img = geoio.GeoImage(self.test_img)
        self.vec = dgsamples.wv2_longmont_1k_vectors.poly_geojson_latlon
        self.badvec = dgsamples.bayou_vectors.poly

    def tearDown(self):
        # Remove gdal image object
        self.img = None
        # Find temp files from test directory and remove
        tmp_files = tt.files.search(os.path.dirname(os.path.realpath(__file__)),
                                    ["*temp*", "*tmp*"], case_sensitive=False)

        print(tmp_files)
        [os.remove(f) for f in tmp_files]

    def test_get_data_noOptions(self):
        for x in self.img.iter_vector(vector=self.vec):
            self.assertIsInstance(x,np.ndarray)

    def test_get_data_noOptions_badVec(self):
        for x in self.img.iter_vector(vector=self.badvec):
            self.assertIsNone(x)

    def test_get_data_propTrue(self):
        for x in self.img.iter_vector(vector=self.vec,properties=True):
            self.assertIsInstance(x,tuple)
            self.assertIsInstance(x[0], np.ndarray)
            self.assertIsInstance(x[1], dict)

    def test_get_data_propTrue_badVec(self):
        for x in self.img.iter_vector(vector=self.badvec,properties=True):
            self.assertIsInstance(x,tuple)
            self.assertIsNone(x[0])
            self.assertIsInstance(x[1],dict)

    def test_get_data_propStr(self):
        for x in self.img.iter_vector(vector=self.vec,properties='testfloat'):
            self.assertIsInstance(x,tuple)
            self.assertIsInstance(x[0], np.ndarray)
            self.assertIsInstance(x[1], dict)
            self.assertIsInstance(x[1]['testfloat'], float)

    def test_get_data_propList(self):
        for x in self.img.iter_vector(vector=self.vec,
                                      properties=['testfloat', 'teststr']):
            self.assertIsInstance(x[0], np.ndarray)
            self.assertIsInstance(x[1], dict)
            self.assertIsInstance(x[1]['teststr'], str)
            self.assertIsInstance(x[1]['testfloat'], float)
            self.assertEqual(len(x[1]), 2)

    def test_get_data_propListBad(self):
        with warnings.catch_warnings(record=True) as w:
            for x in self.img.iter_vector(vector=self.vec,
                                          properties=['testfloatBad']):
                self.assertIsInstance(x[0], np.ndarray)
                self.assertIsNone(x[1])
                self.assertTrue(len(w) == 1)

    def test_get_data_propListGoodBad(self):
        for x in self.img.iter_vector(vector=self.vec,
                                      properties=['testfloatBad','teststr']):
            self.assertIsInstance(x, tuple)
            self.assertIsInstance(x[0], np.ndarray)
            self.assertIsInstance(x[1], dict)
            self.assertIsInstance(x[1]['teststr'], str)
            self.assertRaises('warn.warning')
            self.assertEqual(len(x[1]),1)

    def test_get_data_filtDict(self):
        tmp = [x for x in
               self.img.iter_vector(vector=self.vec, filter={'id':2})]
        self.assertEqual(len(tmp),2)
        self.assertIsNone(tmp[0],None)
        self.assertIsInstance(tmp[1],np.ndarray)

    def test_get_data_filtDictBadKey(self):
        with warnings.catch_warnings(record=True) as w:
            for x in self.img.iter_vector(vector=self.vec, filter={'idbad':2}):
                self.assertIsNone(x)
                self.assertTrue( len(w) == 1 )

    def test_get_data_filtDictBadValue(self):
        for x in self.img.iter_vector(vector=self.vec, filter={'id': 2222}):
            self.assertIsNone(x)

    def test_get_data_filtList(self):
        tmp = [x for x in self.img.iter_vector(vector=self.vec,
                                            filter=[{'id': 2},{'id':1}])]
        self.assertEqual(len(tmp), 2)
        self.assertIsInstance(tmp[0], np.ndarray)
        self.assertIsInstance(tmp[1], np.ndarray)

    def test_get_data_filtListBadValue(self):
        tmp = [x for x in self.img.iter_vector(vector=self.vec,
                                        filter=[{'id': 2222}, {'id': 1}])]
        self.assertEqual(len(tmp), 2)
        self.assertIsInstance(tmp[0], np.ndarray)
        self.assertIsNone(tmp[1])

    def test_get_data_propStr_filtList(self):
        tmp = [x for x in self.img.iter_vector(vector=self.vec,
                        properties='teststr',
                        filter=[{'id': 2},{'id': 1}])]
        self.assertTrue(len(tmp) == 2)
        self.assertIsInstance(tmp[0][0], np.ndarray)
        self.assertTrue(len(tmp[0][1]) == 1)
        self.assertIsInstance(tmp[1][0], np.ndarray)
        self.assertTrue(len(tmp[1][1]) == 1)

    def test_get_data_propList_filtList(self):
        tmp = [x for x in self.img.iter_vector(vector=self.vec,
                          properties=['teststr','testfloat'],
                          filter=[{'id': 2},{'id': 1}])]
        self.assertEqual(len(tmp), 2)
        self.assertIsInstance(tmp[0][0], np.ndarray)
        self.assertTrue(len(tmp[0][1]) == 2)
        self.assertIsInstance(tmp[1][0], np.ndarray)
        self.assertTrue(len(tmp[1][1]) == 2)

class TestDGImage(unittest.TestCase):
    """Testing for :
    DGImage class
    """
    # Run defined method setUp to setup the test environment
    def setUp(self):
        self.test_img = dgsamples.wv2_longmont_1k.ms
        #self.test_img = "../data/imagefiles/053792616010_01/053792616010_01_P001_MUL/14JUN20181517-M2AS-053792616010_01_P001.TIL"
        #self.test_geoimg = "../data/imagefiles/smalldgdelivery/053792616010_01/053792616010_01_P001_MUL/14JUN20181517-M2AS-053792616010_01_P001.TIF"
        #self.test_img_at_sensor_rad = "../data/imagefiles/smalldgdelivery_ms_scaled_to_rad_uW_per_cm2_nm_sr.tif"
        #self.test_img_toa_ref = "../data/imagefiles/smalldgdelivery_ms_scaled_to_toa_ref.tif"
        self.img = geoio.dg.DGImage(self.test_img)
        self.geoimg = geoio.GeoImage(self.test_img)

    def tearDown(self):
        self.img = None
        self.geoimg = None
        # # Remove VRT files from test data dir
        # test_dir = os.path.dirname(os.path.realpath(__file__))
        # search_dir = test_dir+os.sep+os.pardir+os.sep+'data'
        # vrt_files=tt.files.search(search_dir,'*.VRT',
        #                           case_sensitive=False,depth=10)

    def test_DGImage_dg_meta_exists(self):
        self.assertIsInstance(self.img.meta_dg,tt.bunch.OrderedBunch)

    def test_DGImage_dg_meta_abscal(self):
        self.assertTrue(len(self.img.meta.effbandwidth)==8)
        self.assertTrue(len(self.img.meta.abscalfactor)==8)
        self.assertIsInstance(self.img.meta.effbandwidth[0],float)

    def test_DGImage_get_data_as_at_sensor_rad_dtype(self):
        """Output for at sensor radiance should be a numpy array of float32."""
        a = self.img.get_data_as_at_sensor_rad()
        self.assertEqual(a.dtype,np.dtype('float32'))

    def test_DGImage_get_data_as_toa_ref_dtype(self):
        """Output for toa reflectance should be a numpy array of int16 data."""
        a = self.img.get_data_as_toa_ref()
        self.assertEqual(a.dtype,np.dtype('int16'))

    # def test_DGImage_get_data_as_at_sensor_rad_old(self):
    #     """The image loaded into tmp was produced by ENVI."""
    #     a = self.geoimg.get_data()
    #     a = self.img.get_data_as_at_sensor_rad()
    #     tmp = geoio.GeoImage(self.test_img_at_sensor_rad)
    #     b = tmp.get_data()*10  # Convert to W/(m^2*sr*nm)
    #     self.assertIsNone(np.testing.assert_array_almost_equal(a,b,decimal=4))

    def test_DGImage_get_data_as_at_sensor_rad(self):
        """Spot check image against single pixel calculations using the
        following equation:
        L = DN * (ACF/EBW) * (2 - Gain) - Offset
        ACF abscal factor from meta data
        EBW effectiveBandwidth from meta data
        Gain provided by abscal from const
        Offset provided by abscal from const
        """
        a = self.img.get_data_as_at_sensor_rad()
        b = self.img.get_data()
        #b = self.geoimg.get_data()

        # Hard coded from IMD of the test image
        abscalfactor=np.array([0.009295654, 0.01783568,
                               0.01364197, 0.006810718,
                               0.01851735, 0.006063145,
                               0.02050828, 0.009042234])
        effbandwidth=np.array([0.0473, 0.0543, 0.063, 0.0374,
                               0.0574, 0.0393, 0.0989, 0.0996])

        # Dynamic pull from const file that can be updated
        gain=np.array(const.DG_ABSCAL_GAIN['WV02_MULTI'])
        offset=np.array(const.DG_ABSCAL_OFFSET['WV02_MULTI'])

        # Run check for sets of pixel using the simplest per-pixel calc
        for k in range(0,7):
            for i in [100,200,300,400]:
                for j in [100,200,300,400]:
                    pcheck=self.rad_pixel_calc(b,i,j,k,abscalfactor,
                                              effbandwidth,gain,offset)
                    self.assertAlmostEqual(pcheck,a[k,i,j],places=4)

    def rad_pixel_calc(self,data,i,j,k,abscal,effbw,gain,offset):
        # Changed equation Feb. 2016 to match new cal factors from Jan. 2016
        return data[k,i,j]*(abscal[k]/effbw[k])*(gain[k])+offset[k]

    # def test_DGImage_get_data_as_toa_ref_old(self):
    #     """The image loaded into tmp was produced by a running version of my
    #     code with output verified against Python toa_ref code written by
    #     others.
    #     """
    #     a = self.img.get_data_as_toa_ref()
    #     tmp = geoio.GeoImage(self.test_img_toa_ref)
    #     b = tmp.get_data()
    #     self.assertIsNone(np.testing.assert_array_almost_equal(a,b,decimal=6))

    def test_DGImage_get_data_as_toa_ref(self):
        a = self.img.get_data_as_toa_ref()
        b = self.img.get_data()
        #b = self.geoimg.get_data()

        # Hard coded from IMD of the test image
        abscalfactor=np.array([0.009295654, 0.01783568,
                               0.01364197, 0.006810718,
                               0.01851735, 0.006063145,
                               0.02050828, 0.009042234])
        effbandwidth=np.array([0.0473, 0.0543, 0.063, 0.0374,
                               0.0574, 0.0393, 0.0989, 0.0996])
        theta_s=19.20000000000000

        # Dynamic pull from const file that can be updated
        gain=np.array(const.DG_ABSCAL_GAIN['WV02_MULTI'])
        offset=np.array(const.DG_ABSCAL_OFFSET['WV02_MULTI'])

        # Hard coded values for the test image
        e_sun=[1773.81, 2007.27, 1829.62, 1701.85,
               1538.85, 1346.09, 1053.21, 856.599]
        d_es=1.0161415338516235

        # Run check for sets of pixel using the simplest per-pixel calc
        for k in range(0,7):
            for i in [100,200,300,400]:
                for j in [100,200,300,400]:
                    pcheck=self.toa_pixel_calc(b,i,j,k,abscalfactor,
                                              effbandwidth,gain,offset,
                                              e_sun,d_es,theta_s)
                    #print(pcheck)
                    #print(a[k,i,j])
                    self.assertAlmostEqual(pcheck/10000.0,
                                           a[k,i,j]/10000.0,
                                           places=3)

    def toa_pixel_calc(self,data,i,j,k,abscal,effbw,gain,offset,
                       e_sun,d_es,theta_s):
        L=self.rad_pixel_calc(data,i,j,k,abscal,effbw,gain,offset)
        R=(d_es**2*np.pi*L)/(e_sun[k]*np.cos(np.deg2rad(theta_s)))
        return (R*10000).astype('int16')

    def test_DGImage_idl_available(self):
        self.assertTrue(tt.cmd_line.is_tool('idl'))

    def test_DGImage_envi_available(self):
        self.assertTrue(tt.cmd_line.is_tool('envi'))

class TestGeoioFuncs(unittest.TestCase):
    """Testing for :
    Base level functions in Geoio
    """

    def test_request_band_alias_list(self):
        sat_id = 'WV02_MULTI'
        band_alias = ['C','Y']
        out = geoio.dg.get_alias_band_numbers(sat_id, band_alias)
        self.assertEqual(out,[1,4])

    def test_request_band_alias_group(self):
        sat_id = 'WV02_MULTI'
        band_alias = 'VIS'
        out = geoio.dg.get_alias_band_numbers(sat_id, band_alias)
        self.assertEqual(out,[1,2,3,4,5])

    def test_request_band_alias_group2(self):
        sat_id = 'WV02_MULTI'
        band_alias = 'RGBN'
        out = geoio.dg.get_alias_band_numbers(sat_id, band_alias)
        self.assertEqual(out,[5,3,2,7])

    def test_request_band_numbers(self):
        sat_id = 'WV02_MULTI'
        band_alias = [5,4,6]
        out = geoio.dg.get_alias_band_numbers(sat_id, band_alias)
        self.assertEqual(out,[5,4,6])

    def test_request_band_numbers_and_alias(self):
        sat_id = 'WV02_MULTI'
        band_alias = ['C','Y',3]
        out = geoio.dg.get_alias_band_numbers(sat_id, band_alias)
        self.assertEqual(out,[1,4,3])

    def test_request_bad_band_alias_group(self):
        sat_id = 'WV02_MULTI'
        band_alias = 'BADALIAS'
        with self.assertRaises(KeyError):
            geoio.dg.get_alias_band_numbers(sat_id, band_alias)

if __name__ == '__main__':
    unittest.main()


'''
This module contains handler classes for DigitalGlobe imagery.  Includes XML and
PVL (IMD, etc) meta data handling, image date retrieval (UTC to local via
tzwhere), and spectral processing built on top of the GeoImage class
capabilities.
'''


import collections
import datetime
import os
import re
import logging

import ephem
import numpy as np
import pytz
import xmltodict
from pkg_resources import resource_filename
from tzwhere import tzwhere

import tinytools as tt
from base import GeoImage
import constants as const

# Module setup
logger = logging.getLogger(__name__)
# To get access to logging statements from the command line:
# import logging
# logging.basicConfig(level=logging.DEBUG) # or your desired level

class DGImage(GeoImage):
    """ Input can be .TIL, .VRT, OR .TIF.  If .TIL or .VRT, checking is done
    for tiles that belong to the virtual dataset.  If .TIF, then an
    XML or IMD file should exist with the exact same name (sans suffix).  No
    checking is done that the meta file actually belongs with the .TIF.  e.g.
    if an XML file belongs to DN data but surface reflectance data is
    passed, several of this classes methods might explode in bad ways.

    meta data is stored as:
    DGImage.meta_dg [populated by data from DigitalGlobe XML or pvl files -
                     i.e. IMD, RPB, etc.]
    DGImage.meta [populated from GeoImage class __init__]
    DGImage.files
            === Populated by GeoImage ===
            -fin  [file passed in - virtual or otherwise]
            -gdal_file_list  [GetFileList returned from gdal]
            -dfile  [Working target data file (i.e. VRT if TIL passed,
                        otherwise it wil be the same as fin]
            -dfile_tiles  [The tiles of the virtual data set - if the data set
                           isn't virtual, this should be = "files.dfile"]

            === Populated by DGImage ===
            -meta [all meta files, first one is used to populate meta_dg]
            -spectral_data_sets  [Dictionary of Radiance, TOA Reflectance,
                                  and DGAComp as True/False is present/not]
            -dgacomp  [link to DGAComp file]
            -dgacomp_tiles  [The tiles of the virtual data set - if the data set
                            isn't virtual, this should be = "files.dgacomp"]
            -dgacomp_aodmap [link to aop map file generated by dgacomp.  Only
                             present if DGAComp in spectral_data_sets is True]
            -rad  [link to radiance file]
            -rad_tiles  [The tiles of the virtual data set - if the data set
                        isn't virtual, this should be = "files.rad"]
            -toa  [link to toa reflectance file]
            -toa_tiles  [The tiles of the virtual data set - if the data set
                        isn't virtual, this should be = "files.toa"]
    """

    def __init__(self, dg_file_in, derived_dir=None):  #,dg_img_directory
        """Figure out what the main dataset (create a vrt if necessary) and
        set the meta data file/list.
        """
        ### Trigger super __init__ and set the GeoImage stuff
        # ... variables created here:
        # See above for list of "Populated by GeoImage"
        super(DGImage,self).__init__(dg_file_in, derived_dir=derived_dir)
        #########  IMPORTANT !!!!  ##############################
        # DO NOT SUBCLASS WITH THIE SELF REFERENCED SUPER FROM BELOW!
        # INFINITE RECURSION WILL RESULT.  I HAVE THIS IN
        # BECAUSE THIS FIXES AUTORELOAD IN IPYTHON FOR DEVELOPMENT
        # Use the super line above for deployed code
        ##########################################################
        #super(self.__class__,self).__init__(dg_file_in, derived_dir=derived_dir)

        # Get the file name, full file directory, and flist
        ifile = self.files.dfile
        fname = os.path.basename(ifile)
        ifile_base = os.path.splitext(ifile)[0]
        dgbase = fname.split("-")[0]
        fdir = os.path.dirname(ifile)
        flist = os.listdir(fdir)
        flist = [os.path.join(fdir, x) for x in flist]

        ## Populate DG metadata files
        self.files.meta = []
        for e in const.DG_META:
            found_mfiles = tt.files.filter(flist,ifile_base+e,
                                                case_sensitive=False)
            self.files.meta = self.files.meta + found_mfiles

        if not self.files.meta:
            raise ValueError("Failed to find any DigitalGlobe metadata "
                             "files when initializing the DGImage Object.")

        ## Read DG metadata files
        if tt.files.filter(self.files.meta[0], '*.XML', case_sensitive=False):
            # There is an xml file, so lets just use that.
            self._read_dg_dir_meta_xml(self.files.meta[0])
        else:
            # Then there was not an xml so we need to assemble the pvl files.
            self._read_dg_dir_meta_pvl(self.files.meta)

        ## Populate the spectral files for this DGImage
        self._set_dg_spectral_files()

        ## Populate the spectral files for this DGImage
        self._set_dg_meta()

    def _read_dg_dir_meta_xml(self,xml_file):
        # Load the XML dg_meta_file into a dictionary
        with open(xml_file) as fd:
            tmp_dict = xmltodict.parse(fd.read())

        # Filter data types to more easily access members
        tt.bunch.guess_dtype_convert_dict(tmp_dict)

        # Cap everything and filter for stupid pvl to xml differences
        # delivered from the factory
        self._update_dict_differences(tmp_dict)

        # Make a OrderedBunch object for easy command line access
        self.meta_dg = tt.bunch.ordered_bunchify(tmp_dict['ISD'])

    def _read_dg_dir_meta_pvl(self,pvl_file_list):
        # Read dg meta data from the IMD file
        tmp_dict = {'isd':{}}
        add_to_loc = tmp_dict['isd']
        for x in pvl_file_list:
            xend = os.path.splitext(x)[1].split('.')[1].upper()
            add_to_loc[xend]=tt.pvl.read_from_pvl(x)

        # Filter for stupid pvl to xml differences from the factory
        self._update_dict_differences(tmp_dict,'IMAGE_1','IMAGE')

        # Cap everything and filter for stupid pvl to xml differences
        # delivered from the factory
        tt.bunch.guess_dtype_convert_dict(tmp_dict)

        # Make a OrderedBunch object for easy command line access
        self.meta_dg = tt.bunch.ordered_bunchify(tmp_dict['ISD'])

    def _update_dict_differences(self,d,input=None,output=None):
        for x in d.keys():
            if x == input:
                d[output.upper()] = d.pop(x)
                x = output.upper()
            else:
                d[x.upper()] = d.pop(x)
                x = x.upper()

            if isinstance(d[x],dict):
                self._update_dict_differences(d[x],input,output)
            elif isinstance(d[x], collections.OrderedDict):
                self._update_dict_differences(d[x],input,output)
            else:
                pass

    def _set_dg_meta(self):

        # If for some reason the meta object wasn't created, create it
        # This can happen when a class overloads that meta creation routines
        # in the super object.
        try:
            tmp = self.meta
        except AttributeError:
            self.meta = tt.bunch.OrderedBunch()

        self.meta.satid = self.meta_dg.IMD.IMAGE.SATID
        self.meta.bandid = self.meta_dg.IMD.BANDID
        self.meta.catid = self.meta_dg.IMD.IMAGE.CATID

        abscalfactor = []
        effbandwidth = []
        tdilevel = []
        lat = []
        lon = []
        hae = []

        for x in self.meta_dg.IMD:
            if 'BAND_' in x:
                abscalfactor.append(float(self.meta_dg.IMD[x]['ABSCALFACTOR']))
                effbandwidth.append(float(self.meta_dg.IMD[x]['EFFECTIVEBANDWIDTH']))
                tdilevel.append(float(self.meta_dg.IMD[x]['TDILEVEL']))
                lat.append(float(self.meta_dg.IMD[x]['ULLAT']))
                lon.append(float(self.meta_dg.IMD[x]['ULLON']))
                hae.append(float(self.meta_dg.IMD[x]['ULHAE']))

        # Set band data collected from IMD/XML
        self.meta.abscalfactor = abscalfactor
        self.meta.effbandwidth = effbandwidth
        self.meta.tdilevel = tdilevel

        # Set data from constants file
        sat_index = self.meta.satid.upper() + "_" + \
                    self.meta.bandid.upper()
        self.meta.band_names = const.DG_BAND_NAMES[sat_index]
        self.meta.band_centers = const.DG_WEIGHTED_BAND_CENTERS[sat_index]

        # Longitude will be +/-180, Latitude will be +/-90, both with
        # precision of eight decimal places
        # HAE is in meters above WGS84 ellipsoid
        #ToDo - move reported point to the center of the image
        self.meta.latlonhae = (lat[0], lon[0], hae[0])

        # Date/time of collection from EarliestAcqTime
        # Don't use "map_projected_product" time for this to work on 1b files
        # dtstr = self.meta_dg.IMD.MAP_PROJECTED_PRODUCT.EARLIESTACQTIME
        dtstr = self.meta_dg.IMD.IMAGE.FIRSTLINETIME
        time_vars = parse_dg_time_str(dtstr)
        tzinfo_utc = pytz.timezone('UTC')
        self.meta.img_datetime_obj_utc = datetime.datetime(
            *time_vars, tzinfo=tzinfo_utc)

    def _set_dg_spectral_files(self,path=None):
        # Get the search directory
        data_file = self.files.dfile

        if (not path) & hasattr(self,'derived_dir'):
            path=self.derived_dir

        (self.files.rad,self.files.rad_tiles) = \
            self._get_dg_spectral_files(data_file,
                                        const.DG_SPEC['RAD_IMGS'],
                                        path=path)

        (self.files.toa,self.files.toa_tiles) = \
            self._get_dg_spectral_files(data_file,
                                        const.DG_SPEC['TOA_IMGS'],
                                        path=path)

        (self.files.dgacomp,self.files.dgacomp_tiles) = \
            self._get_dg_spectral_files(data_file,
                                        const.DG_SPEC['DGACOMP_IMGS'],
                                        path=path)

        (self.files.dgacomp_aodmap,tmp_dump) = \
            self._get_dg_spectral_files(data_file,
                                        const.DG_SPEC['DGACOMP_AOD'],
                                        path=path)

        ### ToDo - Add DGACOMP_OTHER
        # Find all DGAComp files not in dgacomp or dgacomp_aodmap
        if self.files.dgacomp is not None:
            search_dirs = []
            if path:
                search_dirs.append(path)
            else:
                search_dirs.append(os.path.dirname(data_file))

            for search_dir in search_dirs:
                tmp_files = tt.files.search(search_dir, ['*DG-AComp*', '*DGAComp*'],
                                            depth=2, case_sensitive=False)

            dgacomp_tracked = []
            dgacomp_tracked.append(self.files.dgacomp)
            dgacomp_tracked.extend(self.files.dgacomp_tiles)
            dgacomp_tracked.append(self.files.dgacomp_aodmap)

            [tmp_files.remove(x) for x in dgacomp_tracked]

            self.files.dgacomp_other = tmp_files
        else:
            self.files.dgacomp_other = None


    def _get_dg_spectral_files(self, dfile, search_endings, path=None):

        sestar = [os.path.splitext(os.path.basename(dfile))[0] + x + '.' + '*'
                  for x in search_endings]

        search_dirs = []
        if path:
            search_dirs.append(path)
        else:
            search_dirs.append(os.path.dirname(dfile))

        for search_dir in search_dirs:
            tmp_files = tt.files.search(search_dir, sestar, depth=2)

        if not tmp_files:
            return (None,[])

        if not (len(tmp_files) <= 2):
            raise ValueError("There should be no more than two files "
                             "returned for the supported file types (TIL, "
                             "VRT, TIF, ENVI).")
        if len(tmp_files) == 2:
            tif_file = tt.files.filter(tmp_files, '*.TIF')
            if tif_file:
                tmp_files.remove(tif_file[0])

        return self._get_file_and_tiles(tmp_files[0])

    def populate_img_datetime_obj_local(self,w=None):
        ## Get the time/date at the image collect location
        logger.debug('')
        logger.debug('Adding self.img_datetime_obj_local to obj data members.')
        logger.debug('')
        try:
            self.img_tz = w.tzNameAt(self.meta.latlonhae[0],
                                     self.meta.latlonhae[1])
        except:
            ###
            # Loading a tzwhere intstance from scratch...
            # This could be theoretically changed by including a '
            # timezone database or a module that wraps one instead of '
            # the json file.'
            ###
            #logger.debug('')
            logger.debug('Opening a tzwhere instance for access to the json file.  '
                  'This is slow!  You can pass in a pregenerated instance '
                  'of the json database to the variable w in order to speed '
                  'up the processing of multiple images.  This is done by '
                  'importing tzwhere using "from tzwhere import tzwhere" and '
                  'then creating the instance using "w=tzwhere.tzwhere()"  '
                  'See notes in code for possible development to improve '
                  'speed if needed.')
            logger.debug('')
            w = tzwhere.tzwhere()
            self.meta.img_tz = w.tzNameAt(
                                            self.meta.latlonhae[0],
                                            self.meta.latlonhae[1])
            if not self.meta.img_tz:
                logger.debug('')
                logger.debug('No time zone code returned, this location may be in '
                      'the middle of the ocean.  Defaulting to a simple '
                      'latitude based time zone calculation.')
                offset = self._calc_gmtoffset(self.meta.latlonhae[1])
                #print offset
                self.meta.img_tz = 'Etc/GMT' + str(offset)

        tzinfo_img = pytz.timezone(self.meta.img_tz)
        self.meta.img_datetime_obj_local = \
            self.meta.img_datetime_obj_utc.astimezone(tzinfo_img)

    def _calc_gmtoffset(self,l):
        """Calculate the general time zone offset based on latitude alone."""
        offset=int(((l+180+7.5)//15)-12)
        # There is no such thing as a -12 time zone offset.  By the math,
        # the difference offsets go from -12.5 to 12.5, but by convention,
        # the fractional offsets are convereted to 12.  So, the offset range
        # is [-11,12] inclusive.  There are areas that use timezones of +13
        # and +14 in the pacific, but these should be taken care of by the
        # tzwhere call above.
        if offset==-12:
            return 12
        else:
            return offset

    # ToDo - change to passing **kwargs to super.  Before that happens, shoudl
    # probably figure out how to merge the docstring from super with below.
    def get_data(self, component = None,
                       bands = None,
                       window = None,
                       buffer = None,
                       geom=None,
                       mask = False,
                       mask_all_touched=False,
                       virtual = False,
                       stype = None):
        """Get image data with ability to output a data frame or request
        keyword arguments to the parent get_data function.  These include
        requesting certain bands and specific components.  This function will
        hijack any bands request and look to see if the request matches
        band aliases defined in const before passing a converted list of
        ints to super."""

        # Set spectral retrival if requested
        if stype:
            raise NotImplementedError("stype handling in get_data is not " \
                                      "implemented yet.  Please use the " \
                                      "dedicated methods.")

        # Set satelite index to query const dictionaries
        # Set initial band names that might be updated below
        sat_index = self.meta.satid.upper() + "_" + \
                    self.meta.bandid.upper()
        bnames = const.DG_BAND_NAMES[sat_index]

        # handle options request for individual bands
        if bands:
            band_nums = get_alias_band_numbers(sat_index,bands)
            band_nums = [x for x in band_nums if x is not None]
            bnames = [bnames[x-1] for x in band_nums if x is not None]
        else:
            band_nums = get_alias_band_numbers(sat_index,bnames)

        if len(band_nums) is 0:
            raise ValueError("No band values were found in the requested " \
                             "alias.")

        # Call super with keywords passed in and/or convereted above
        data = super(self.__class__,self).get_data(component = component,
                                           bands = band_nums,
                                           window = window,
                                           buffer = buffer,
                                           geom=geom,
                                           mask = mask,
                                           mask_all_touched=mask_all_touched,
                                           virtual = virtual)

        return data

    def get_data_as_at_sensor_rad(self,component=None):
        """Read data from sensor as at sensor radiance.  The returned values
        are in W/(m^2*sr*nm).  The values are calculated with known
        gain/offset values pull from the geoio.constants file as provided by
        the DigitalGlobe calibration team.  The absolute calibration factor
        and effective bandwidths are pull from the IMD files for each image."""

        # Set satellite index to look up cal factors
        sat_index = self.meta.satid.upper() + "_" + \
                    self.meta.bandid.upper()

        # Set scale for at sensor radiance
        # Eq is:
        # L = DN * (ACF/EBW) * (2 - Gain) - Offset
        # Above equation is retired as of Jan. 2016, the equation to use with
        # the most recent cal coefficients is:
        # L = GAIN * DN * (ACF/EBW) + Offset
        # ACF abscal factor from meta data
        # EBW effectiveBandwidth from meta data
        # Gain provided by abscal from const
        # Offset provided by abscal from const
        num = np.asarray(self.meta.abscalfactor) # Should be nbands length
        den = np.asarray(self.meta.effbandwidth)  # Should be nbands length
        gain = np.asarray(const.DG_ABSCAL_GAIN[sat_index])
        scale = (num/den)*(gain)

        # Shape for easy multiple
        scale = scale[:, np.newaxis, np.newaxis]
        offset = np.asarray(const.DG_ABSCAL_OFFSET[sat_index])
        offset = offset[:, np.newaxis, np.newaxis]

        # Pull raw data
        if component is not None:
            data = self.get_data(component=component)
        else:
            data = self.get_data()

        # Return scaled data
        out = (data*scale.astype('float32'))+offset.astype('float32')

        return out

    def get_data_as_toa_ref(self,component=None):
        """Get data in a numpy array as top-of-atmosphere reflectance.
        Output is in scaled reflectance 0-10,000.
        Input:  data (numpy array in bands,lines,samples)
                absCalFactor (int or list of nbands)
                effBandWidth (int or list of nbands)
                l_band (list of nbands)
                e_sun (list of nbands)
                d_es (int/float)
                theta_s (int/float)
        Output: image data in TOA reflectance
        """

        # Set satellite index to look up cal factors
        sat_index = self.meta.satid.upper() + "_" + \
                    self.meta.bandid.upper()

        # Set scale for at sensor radiance
        # Eq is:
        # L = DN * (ACF/EBW) * (2 - Gain) - Offset
        # Above equation is retired as of Jan. 2016, the equation to use with
        # the most recent cal coefficients is:
        # L = GAIN * DN * (ACF/EBW) + Offset
        # ACF abscal factor from meta data
        # EBW effectiveBandwidth from meta data
        # Gain provided by abscal from const
        # Offset provided by abscal from const
        num = np.asarray(self.meta.abscalfactor) # Should be nbands length
        den = np.asarray(self.meta.effbandwidth)  # Should be nbands length
        gain = np.asarray(const.DG_ABSCAL_GAIN[sat_index])
        scale = (num/den)*(gain)

        # Shape for easy multiple
        scale = scale[:, np.newaxis, np.newaxis]
        offset = np.asarray(const.DG_ABSCAL_OFFSET[sat_index])
        offset = offset[:, np.newaxis, np.newaxis]

        ## Read Esun for the bands of this satellite from the const file
        e_sun_index = self.meta.satid.upper() + "_" + \
                      self.meta.bandid.upper()
        e_sun = const.DG_ESUN[e_sun_index]

        ## Calculate the generic earth_sun distance
        # For a test image over Longmont, the code below produces:
        # 1.0161819458007812
        # sun = ephem.Sun()
        # sun.compute(self.img_datetime_obj)
        # d_es = sun.earth_distance

        ## Calculate the earth_sun distance for the image location
        # This is probably overkill, but... whatever, it was easy.  =)
        # For a test image over Longmont, the code below produces:
        # 1.0161415338516235
        sun = ephem.Sun()
        img_obs = ephem.Observer()
        img_obs.lon = self.meta.latlonhae[1]
        img_obs.lat = self.meta.latlonhae[0]
        img_obs.elevation = self.meta.latlonhae[2]
        img_obs.date = self.meta.img_datetime_obj_utc
        sun.compute(img_obs)
        d_es = sun.earth_distance

        ## Pull sun elevation from the image metadata
        #theta_s can be zenith or elevation - the calc below will us either
        # a cos or s in respectively
        #theta_s = float(self.meta_dg.IMD.IMAGE.MEANSUNEL)
        theta_s = 90-float(self.meta_dg.IMD.IMAGE.MEANSUNEL)

        # Set up the solar and geometry parameters
        d = np.repeat(d_es, self.shape[0]) # earth-sun distance
        e = np.asarray(e_sun)          # e_sun for each band
        te = np.repeat(theta_s, self.shape[0])  # sun elevation

        # Perform the TOA reflectance calculation
        #scale2 = (d**2*np.pi)/(e*np.sin(np.deg2rad(te)))
        scale2 = (d ** 2 * np.pi) / (e * np.cos(np.deg2rad(te)))

        # Shape for easy multiple
        scale2 = scale2[:, np.newaxis, np.newaxis]

        # Pull raw data
        if component is not None:
            data = self.get_data(component=component)
        else:
            data = self.get_data()

        # Set data types for output
        # scale = scale.astype('float32')
        # offset = offset.astype('float32')
        # scale2 = scale2.astype('float32')

        # Set data types for output
        scale = scale.astype('float16')
        offset = offset.astype('float16')
        scale2 = scale2.astype('float16')

        # Return scaled data
        return ((((data*scale)+offset)*scale2)*10000).astype('int16')

    def create_at_sensor_rad_files(self,path=None,components=True):
        if hasattr(self,'derived_dir'):
            path=self.derived_dir
        #Convert to files, each component if requested.
        if not components:
            # Create single file from image data
            fl = list(os.path.splitext(self.files.dfile))
            if path:
                bn = os.path.basename(fl[0])
                fl[0] = os.path.join(path, bn)
            new_fname = fl[0] + const.DG_SPEC['RAD_IMGS'][0] + fl[1]
            data = self.get_data_as_at_sensor_rad()
            self.write_img_like_this(new_fname,data)
            # Add to dg_img_rad
            self.dg_img_rad = new_fname
        elif components:
            if (self.files.dfile_tiles[0] == self.files.dfile):
                logger.debug("This data set does not appear to have "
                      "componenets, reverting to the main file.")
                self.create_at_sensor_rad_files(components=False,path=path)
            else:
                # Create each component file
                flist_for_vrt = []
                for yi,yv in enumerate(self.files.dfile_tiles):
                    x = GeoImage(yv)
                    fl = list(os.path.splitext(x.files.dfile))
                    if path:
                        bn = os.path.basename(fl[0])
                        fl[0] = os.path.join(path, bn)
                    new_fname = fl[0] + const.DG_SPEC['RAD_IMGS'][0] + fl[1]
                    flist_for_vrt.append(new_fname)
                    data = self.get_data_as_at_sensor_rad(component=yi+1)
                    x.write_img_like_this(new_fname,data)
                    x=None
                # Create the vrt
                tmp = list(os.path.splitext(self.files.dfile))
                if path:
                    bn = os.path.basename(tmp[0])
                    tmp[0] = os.path.join(path, bn)
                #Until geoio can write a .TIL file this is commented
                #vrt_name = tmp[0]+const.DG_SPEC['RAD_IMGS'][0]+tmp[1]
                vrt_name = tmp[0] + const.DG_SPEC['RAD_IMGS'][0] + '.VRT'
                cmd = []
                cmd.append("gdalbuildvrt")
                cmd.append(vrt_name)
                for i in flist_for_vrt: cmd.append(i)
                tt.cmd_line.exec_cmd(cmd)
                if not os.path.isfile(vrt_name):
                   raise StandardError("Creation of file "+vrt_name+" "
                                       "failed. This could possibly be a "
                                       "write access problem?")

        # Update the DG spectral file meta data
        self._set_dg_spectral_files()

    def create_toa_ref_files(self,path=None,components=True):
        if hasattr(self,'derived_dir'):
            path=self.derived_dir
        #Convert to files, each component if requested.
        if not components:
            # Create single file from image data
            fl = list(os.path.splitext(self.files.dfile))
            if path:
                bn = os.path.basename(fl[0])
                fl[0] = os.path.join(path, bn)
            new_fname = fl[0] + const.DG_SPEC['TOA_IMGS'][0] + fl[1]
            data = self.get_data_as_toa_ref()
            self.write_img_like_this(new_fname,data)
            # Add to dg_img_rad
            self.dg_img_rad = new_fname
        elif components:
            if (self.files.dfile_tiles[0] == self.files.dfile):
                logger.debug("This data set does not appear to have "
                      "componenets, reverting to the main file.")
                self.create_toa_ref_files(components=False,path=path)
            else:
                # Create each component file
                flist_for_vrt = []
                for yi,yv in enumerate(self.files.dfile_tiles):
                    x = GeoImage(yv)
                    fl = list(os.path.splitext(x.files.dfile))
                    if path:
                        bn = os.path.basename(fl[0])
                        fl[0] = os.path.join(path, bn)
                    new_fname = fl[0] + const.DG_SPEC['TOA_IMGS'][0] + fl[1]
                    flist_for_vrt.append(new_fname)
                    data = self.get_data_as_toa_ref(component=yi+1)
                    x.write_img_like_this(new_fname,data)
                    x=None
                # Create the vrt
                tmp = list(os.path.splitext(self.files.dfile))
                if path:
                    bn = os.path.basename(tmp[0])
                    tmp[0] = os.path.join(path, bn)
                #Until geoio can write a .TIL file this is commented
                #vrt_name = tmp[0]+const.DG_SPEC['TOA_IMGS'][0]+tmp[1]
                vrt_name = tmp[0] + const.DG_SPEC['TOA_IMGS'][0] + '.VRT'
                cmd = []
                cmd.append("gdalbuildvrt")
                cmd.append(vrt_name)
                for i in flist_for_vrt: cmd.append(i)
                tt.cmd_line.exec_cmd(cmd)
                if not os.path.isfile(vrt_name):
                   raise StandardError("Creation of file "+vrt_name+" "
                                       "failed. This could possibly be a "
                                       "write access problem?")

        # Update the DG spectral file meta data
        self._set_dg_spectral_files()

    def create_dgacomp_ref_files(self,path=None,ms_aod_map=None,
                                 force_create=False):
        """Create surface reflectance files using DGAComp.  This currently
        uses the IDL code.

        ms_aod_map must be pass if this is a SWIR or a PAN image.
        """

        if force_create:
            pass
        elif self.files.dgacomp:
            logger.debug("DGAcomp files alearead exist... skipping recreation.")
            return

        # Check that the request is not a .VRT file - I don't think dgacomp
        # can handle that
        if not ((os.path.splitext(self.files.dfile)[-1] != '.VRT') and \
                (os.path.splitext(self.files.dfile)[-1] != '.vrt')):
            raise ValueError("DGAComp doesn't handle VRT files.  Can you "
                             "pass in the TIL file?")

        if (not path) & hasattr(self,'derived_dir'):
            path=self.derived_dir

        # Catch the approriate call for each spectral type
        if self.meta_dg.IMD.BANDID == const.DG_BANDID['IMDXML'][0]: #PAN
            if not ms_aod_map: raise ValueError("An ms_aod_map must be passed "
                                                "for PAN data.")
            inFile = self.files.dfile
            tmp = os.path.splitext(self.files.dfile)
            if path:
                x = os.path.join(path, os.path.basename(tmp[0]))
            else:
                x = tmp[0]
            outFile = x + const.DG_SPEC['DGACOMP_IMGS'][0]
            aodFile = ms_aod_map
        elif self.meta_dg.IMD.BANDID == const.DG_BANDID['IMDXML'][1]: #MS
            inFile = self.files.dfile
            tmp = os.path.splitext(self.files.dfile)
            if path:
                x = os.path.join(path, os.path.basename(tmp[0]))
            else:
                x = tmp[0]
            outFile = x + const.DG_SPEC['DGACOMP_IMGS'][0]
        elif self.meta_dg.IMD.BANDID == const.DG_BANDID['IMDXML'][2]: #SWIR
            if not ms_aod_map: raise ValueError("An ms_aod_map must be "
                                                "passed for SWIR data.")
            inFile = self.files.dfile
            tmp = os.path.splitext(self.files.dfile)
            if path:
                x = os.path.join(path, os.path.basename(tmp[0]))
            else:
                x = tmp[0]
            outFile = x + const.DG_SPEC['DGACOMP_IMGS'][0]
            aodFile = ms_aod_map

        # This works but uses hard coded file names which is evidentally
        # fragile under certain python deployment circumstances.
        # dgacomp_wrapper = os.path.join(os.path.dirname(__file__),
        #                                'nwl_dgacomp_batch.pro')
        # Running the above with setuptools' pkg_resources
        dgacomp_wrapper = resource_filename(__name__, 'nwl_dgacomp_batch.pro')
        os.path.isfile(dgacomp_wrapper)
        logger.debug(dgacomp_wrapper)
        if not os.path.isfile(dgacomp_wrapper):
            raise ValueError("The DGAComp wrapper files isn't present "
                             "in the module.")

        # Create command line idl call like:
        # idl -e ".run nwl_batch.pro" -args '/path/to/in_file'
        #                                   '/path/to/out_file'
        #                                   '/path/to/ms_aod_map'
        # pass ms_aod_map if in_file is swir or pan

        cmd = []
        cmd.append('idl -e ".run '+dgacomp_wrapper+'" -args')
        cmd.append(const.DGACOMP['TARGET'][const.DGACOMP_INDEX])
        cmd.append(inFile)
        cmd.append(outFile)
        try:
            cmd.append(aodFile)
        except:
            pass

        # Hack subprocess to get aliases into the shell
        # http://stackoverflow.com/questions/25099895/from-python-start-a-shell-that-can-interpret-functions-and-aliases
        # http://stackoverflow.com/questions/12060863/python-subprocess-call-a-bash-alias
        tt.cmd_line.exec_cmd([os.getenv('SHELL'), '-i', '-c', ':;' + ' '.join(cmd)])

        # Add to dg_img_dgacomp
        # Since outFile doesn't have the ending on, I'm search for the file
        # to make sure I find a file that really exists with the correct
        # ending.
        # if os.path.isfile(outFile+'.TIL'):
        #     self.files.dgacomp = outFile+'.TIL'
        # elif os.path.isfile(outfile):
        #     self.files.dgacomp = outFile
        #
        # if os.path.isfile(outFile+const.DG_SPEC['DGACOMP_AOD'][0]):
        #     self.files.dgacomp_aodmap = outFile+const.DG_SPEC['DGACOMP_AOD'][0]

        # Update the DG spectral file meta data
        self._set_dg_spectral_files()

    def delete_rad_files(self,test_only=True):
        """Delete from disk the radiance files that this class can generate.
        If test_only is set to true (default), this will only simulated the
        removal until you pass test_only=False.
        """

        # Empty line for readability
        logger.info('')
        if test_only:
            logger.info("*** NO RADIANCE FILE OPERATIONS WILL BE PERFORMED ***")
            logger.info("*** test_only is set to True (default) ***")
            logger.info("")

        # Find spectral files
        if self.files.rad in self.files.rad_tiles:
            rall = self.files.rad_tiles
        else:
            rall = [self.files.rad]+self.files.rad_tiles

        # Kick out if no files of this spectral type are present
        if rall[0] == None:
            logger.info("No radiance files to be deleted.")
            return

        # Annouce the files to be removed
        logger.info("Deleting radiance files:")
        for f in rall:  logger.info(f)

        if test_only:
            return

        # Remove the files
        [os.remove(f) for f in rall]

        # Redo the spectral file dict and bunch
        self._set_dg_spectral_files()

    def delete_toa_ref_files(self,test_only=True):
        """Delete from disk the toa_reflectance files that this class can
        generate.  If test_only is set to true (default), this will only
        simulated the removal until you pass test_only=False.
        """

        # Empty line for readability
        logger.info('')
        if test_only:
            logger.info("*** NO TOA REFLECTANCE FILE OPERATIONS WILL BE PERFORMED ***")
            logger.info("*** test_only is set to True (default) ***")
            logger.info("")

        # Find spectral files
        if self.files.toa in self.files.toa_tiles:
            tall = self.files.toa_tiles
        else:
            tall = [self.files.toa]+self.files.toa_tiles

        # Kick out if no files of this spectral type are present
        if tall[0] == None:
            logger.info("No TOA files to be deleted.")
            return

        # Print files to be operated on
        logger.info("Deleting TOA reflectance files:")
        for f in tall:  logger.info(f)

        if test_only:
            return

        # Remove the files
        [os.remove(f) for f in tall]

        # Redo the spectral file dict and bunch
        self._set_dg_spectral_files()

    def delete_dgacomp_files(self,test_only=True):
        """Delete from disk the DGAComp files that this class can generate.
        If test_only is set to true (default), this will only simulated the
        removal until you pass test_only=False.
        """

        # Empty line for readability
        logger.info('')
        if test_only:
            logger.info("*** NO DGACOMP FILE OPERATIONS WILL BE PERFORMED ***")
            logger.info("*** test_only is set to True (default) ***")
            logger.info("")

        # Find spectral files
        dall = []
        if self.files.dgacomp in self.files.dgacomp_tiles:
            if self.files.dgacomp_tiles: dall=dall+self.files.dgacomp_tiles
            if self.files.dgacomp_aodmap: dall=dall+[self.files.dgacomp_aodmap]
            if self.files.dgacomp_other: dall=dall+self.files.dgacomp_other
        else:
            if self.files.dgacomp: dall=dall+[self.files.dgacomp]
            if self.files.dgacomp_tiles: dall=dall+self.files.dgacomp_tiles
            if self.files.dgacomp_aodmap: dall=dall+[self.files.dgacomp_aodmap]
            if self.files.dgacomp_other: dall=dall+self.files.dgacomp_other

        # Kick out if no files of this spectral type are present
        if not dall:
            logger.info("No DGAComp files to be deleted.")
            return

        # Print files to be operated on
        logger.info("Deleting DGAComp files:")
        for f in dall:  logger.info(f)

        if test_only:
            return

        # Remove the files
        [os.remove(f) for f in dall]

        # Redo the spectral file dict and bunch
        self._set_dg_spectral_files()

    def delete_all_spectral_files(self,test_only=True):
        """Delete from disk all spectrally derived files that this class can
        generate.  If test_only is set to true (default), this will only
        simulated the removal until you pass test_only=False.
        """
        self.delete_rad_files(test_only=test_only)
        self.delete_toa_ref_files(test_only=test_only)
        self.delete_dgacomp_files(test_only=test_only)

    def get_data_as_surf_ref(self):
        """Get data from DGAcomp files, if they exist."""
        if not self.files.dgacomp:
            raise ValueError('DGAComp files must be precomputed to use ' \
                             'this method.')
        tmp = GeoImage(self.files.dgacomp)
        data = tmp.get_data()
        del tmp

        return data

def parse_dg_time_str(dtstr):
    if not dtstr.endswith('Z'):
        logger.debug(dtstr)
        raise ValueError('The date/time string retrieved from the DG '
                         'meta data file is not of the format expected.')
    # Example:  '2013-02-16T18:28:30.183342Z'
    tmp = re.split("-|T|:|\.|Z", dtstr)
    year = int(tmp[0])
    month = int(tmp[1])
    day = int(tmp[2])
    hour = int(tmp[3])
    min = int(tmp[4])
    sec = int(tmp[5])
    usec = int(tmp[6]) #Fraction second to msecs

    if len(tmp[6]) != 6: raise ValueError("This routine expects there to be "
                                          "six digits in the usec field.")

    return [year,month,day,hour,min,sec,usec]


class DGImageSet(object):
    """This class is designed to encapsulate the imagery from a DigitalGlobe
    delivery directory.
    """

    def __init__(self,list_of_dg_imgs):
        # Initialize the DGImageDir Objects for this image
        self._set_delivery_dir_components(list_of_dg_imgs)

    def _set_delivery_dir_components(self,paths):
        counts = [0,0,0,0]
        self.obj_list = []
        for x in paths:
            y = DGImage(x)
            if y.bandid == 'P':
                self.pan = DGImage(x)
                self.obj_list.append(self.pan)
                counts[0] += 1
            if y.bandid == 'Multi':
                self.mul = DGImage(x)
                self.obj_list.append(self.mul)
                counts[1] += 1
            if y.bandid == 'All-S':
                self.swr = DGImage(x)
                self.obj_list.append(self.swr)
                counts[2] += 1
            if y.bandid == 'All-C':
                self.cavis = DGImage(x)
                self.obj_list.append(self.cavis)
                counts[3] += 1
            y = None

        for x in counts:
            if x>1:
                raise ValueError("More than one of a single image type was "
                                 "passed into DGDelivDir.")

    def __repr__(self):
        """overload repr to print somethign a bit more useful."""
        sss = ''
        sss+=str(self.__class__.__name__)

        # Get components present and file_names
        components = []
        cdirs = []
        try:
            cdirs.append(self.pan.meta.file_name)
            components.append("Pan")
        except: pass
        try:
            cdirs.append(self.mul.meta.file_name)
            components.append("VNIR")
        except: pass
        try:
            cdirs.append(self.swr.meta.file_name)
            components.append("SWIR")
        except: pass
        try:
            cdirs.append(self.cavis.meta.file_name)
            components.append("CAVIS")
        except: pass

        # Add components to string
        sss+=' ('+', '.join(components)+')'

        # Add most common directory to string
        sss+=" [" + os.path.commonprefix(cdirs).split(os.sep)[-2] + "]"

        return sss

    def build_supercube(self):
        logger.error("not implemented yet")
        pass

    def run_dg_acomp(self):
        logger.error("not implemented yet")
        pass


def create_DGDelivDir_objs(spath,stupid_search=True):
    """This function scans for multiple DigitalGlobe image
    delivery directories and attempts to reconstruct them into
    DGDelivDir objects with the correct components (MUL, SWIR, etc.).

    stupid_search is there to help find SWIR and CAVIS orders that don't
    have the same product id.  This is complicated because as of now (Jan.
    3rd, 2015) the factory is not providing CAVIS and SWIR imagery in a
    single order.  So this function has to go search for them based on the
    catid and do manual checks against the image properties (projection,
    etc.) to guess if they belong to together.
    """
    img_exts = []
    for x in list_bandid_dir:
        img_exts.append('*'+x+'*')

    # Find all directories that match the image filter strings above
    all_img_dirs = tt.files.search(spath, img_exts, ret_files=False,
                                   ret_dirs=True, depth=0)

    ## Put together dir and img value lists
    # If the directory does load, it isn't a DG image directory.
    # Also populated info to filter with.
    img_dirs = []   #Final list of image dirs to use
    img_pid = []    #Image product id
    #dir_pid = []   #Directory product id
    img_cid = []    #Image catalog id
    img_bid = []    #Image band id (PAN, MUL, etc)
    img_outf = []   #Image output foramt
    img_plvl = []   #Image product level
    img_scnd = []   #Scan direction
    img_satid = []  #Satellite ID
    img_rker = []   #Resampling kernel
    img_gentm = []  #Product generation time
    for x in all_img_dirs:
        try:
            y = DGImageDir(x)
        except:
            # If an error from above, then this isn't a DG image dir.
            continue
        img_dirs.append(x)
        img_pid.append(y.meta_dg.IMD.PRODUCTORDERID)
        #dir_pid.append((os.path.basename(x).rsplit('_',1))[0])
        img_cid.append(y.meta_dg.IMD.IMAGE.CATID)
        img_bid.append(y.meta_dg.IMD.BANDID)
        img_outf.append(y.meta_dg.IMD.OUTPUTFORMAT)
        img_plvl.append(y.meta_dg.IMD.PRODUCTLEVEL)
        img_scnd.append(y.meta_dg.IMD.IMAGE.SCANDIRECTION)
        img_satid.append(y.meta_dg.IMD.IMAGE.SATID)
        img_rker.append(y.meta_dg.IMD.IMAGE.RESAMPLINGKERNEL)
        img_gentm.append(y.meta_dg.IMD.GENERATIONTIME)
        y = None

    ## Put together image sets
    if not stupid_search:
        # If I don't need to do anything fancy, pass based on product id
        (set_pid,index_pid) = np.unique(img_pid, return_inverse=True)
        index_master = index_pid
    elif stupid_search:
        # Since stupid search is turned on (until the factory starts
        # delivering everything together), start putting the directories
        # together based on related information and a little bit of luck.
        # Checking catids and other values that indicate image directories
        # below to the same product.
        img_cid2 = [x[:3]+x[4:] for x in img_cid]
        (set_cid2,index_cid2) = np.unique(img_cid2, return_inverse=True)
        index_master = index_cid2

        checks=[]
        for x in set(index_master):
            check_bool = np.asarray(index_master == x)
            checks.append(set(np.asarray(img_outf)[check_bool]))
            checks.append(set(np.asarray(img_plvl)[check_bool]))
            checks.append(set(np.asarray(img_scnd)[check_bool]))
            checks.append(set(np.asarray(img_satid)[check_bool]))
            checks.append(set(np.asarray(img_rker)[check_bool]))

        #(set_outf,index_outf) = np.unique(img_outf,return_inverse=True)
        #(set_plvl,index_plvl) = np.unique(img_plvl,return_inverse=True)
        #(set_scnd,index_scnd) = np.unique(img_scnd,return_inverse=True)
        #(set_satid,index_satid) = np.unique(img_satid,return_inverse=True)
        #(set_rker,index_rker) = np.unique(img_rker,return_inverse=True)
        index_master = _split_master(img_outf,index_master)
        index_master = _split_master(img_plvl,index_master)
        index_master = _split_master(img_scnd,index_master)
        index_master = _split_master(img_satid,index_master)
        index_master = _split_master(img_rker,index_master)

        (set_gentm,index_gentm) = np.unique(img_gentm, return_inverse=True)
        logger.debug(img_dirs)

        # TODO --- actually code up this time check
        # Check that the image generation time is +/- one hour
        # Not the best check because gen time is not guarantee
        time_vars = parse_dg_time_str(img_gentm[0])
        tzinfo_utc = pytz.timezone('UTC')
        gentime = datetime.datetime(*time_vars, tzinfo=tzinfo_utc)
        range = datetime.timedelta(hours=1)
        gentime-range<=gentime<=gentime+range

    # Check that there isn't more than one of each band type in the master list
    # TODO --- code up this band array check
    (set_bid,index_bid) = np.unique(img_bid, return_inverse=True)

    ### Generate the DGDelivDir objects
    list_of_objs = []
    for x in set(index_master):
        pass_list=np.asarray(img_dirs)[index_master == x]
        logger.debug(pass_list)
        logger.debug('')
        list_of_objs.append(DGDelivDir(pass_list))

    return list_of_objs


def _split_master(p,master):
    # For each item in master, check if those items are split in p.  If they
    # are, split master by the p split.
    mbool = [True]*len(master)
    for x in set(master):
        ind = [yi for yi,yv in enumerate(master) if yv == x]
        mbool = [False for qi,qv in enumerate(mbool) if master[qi]==x]
        #ind = np.where(master==x)
        #(sm,im) = np.unique(master[ind],return_inverse=True)
        p_atind = [yv for yi,yv in enumerate(p) if yi in ind]
        p_atind_set = set(p_atind)
        if len(p_atind_set) > 1:
            # Need to split index_master
            #(sp,ip) = np.unique(p_atind,return_inverse=True)
            skip = True
            imax = max(master)
            for y in p_atind_set:
                if skip==True:
                    skip=False
                    continue

                indy = [zi for zi,zv in enumerate(p_atind) if zv == y]
                #master = [imax if ]
                master = [iup for ai,av in enumerate(master) if ai]
                indy = [z for z in p[ind] if z == y]
                imax += 1

    return master


def get_alias_band_numbers(dg_band_name,requested_alias):
    """Get the band numbers, for a specific DG image sensor, that correspond to
    a spectral region alias.  If a band is requested by an alias that isn't
    present in the requested band names, a None value is returned.  The
    returned values are base 1 so that they can be passed directly to
    gdal GetRasterBand.
    """

    # Look up band names given the provided dg_band string
    names = const.DG_BAND_NAMES[dg_band_name]

    try:
        alias = const.DG_BAND_ALIASES[requested_alias]
    except TypeError:
        alias = requested_alias
    except KeyError:
        if requested_alias in names:
            alias = requested_alias
        else:
            raise

    # Return the band numbers for the requested alias
    out=[None]*len(alias)
    for i,v in enumerate(alias):
        try:
            if v.upper() in names:
                out[i] = names.index(v.upper())+1
        except:
            out[i] = v
    # The above code replaces the following list comprehension so that the
    # function will pass through integers
    #out=[names.index(v.upper()) if v.upper() in names else None for v in alias]


    return out



#!/usr/bin/env python

import argparse
import os

import geoio.dg
import tinytools as tt
import geoio


def setup():
    ### parse command line input ###
    descrip = """Create the spectral conversion files requested by the arguments."""

    parser = argparse.ArgumentParser(description=descrip)
    parser.add_argument("-input", "--input",
                        required=True,
                        nargs='+',
                        help="A list of file strings to operate on.")
    parser.add_argument("-spectral_ops", "--spectral_ops",
                        required=True,
                        choices=['radiance', 'toa', 'DGAComp'],
                        nargs='+',
                        help="Which spectral operations to perform on the "
                             "input files.")
    parser.add_argument("-outdir", "--outdir",
                        required=False,
                        nargs=1,
                        help="*** NOT IMPLEMENTED YET ***"
                             "Location to write the output files if not the "
                             "default rename at the input location.")
    parser.add_argument("-test_only", "--test_only",
                        required=False,
                        action="store_true",
                        default=False,
                        help="This flag will turn off all spectral operations "
                             "and will return the files that would have "
                             "been operated on if the flag was not passed.")
    parser.add_argument("-clear_all_spectral_files",
                        "--clear_all_spectral_files",
                        required=False,
                        action="store_true",
                        default=False,
                        help="This will delete ALL the spectral files "
                             "associated with "
                             "the files passed in.  This will short circuit"
                             "all spectral operations.  It is STRONGLY "
                             "suggested "
                             "that a test_only check is performed before "
                             "running to prevent unwanted data removal.")

    return parser.parse_args()

    assert (parser.spectral_ops or parser.clear_all_spectral_files), \
        "Either a spectral operation of a deletion of files needs to " \
        "be requested."


def main(args):
    ### Input files ###
    flist = args.input

    ### Output Run Info ###
    # Print the files that are going to be manipulated and, if test_only, exit
    print('')
    print("The input files are:")
    for x in flist:
        print(x)

    ### Validation ###
    # Verify the files all open correctly - you wouldn't want to have stuff crash
    # part of the way through a long run.  So, this will proivde a bit
    # of validation.
    print('')
    print("Verifying the files have the appropriate meta data...")
    for x in flist:
        i1 = geoio.dg.DGImage(x)
        assert i1.meta_dg.IMD
        assert i1.meta.abscalfactor
        assert i1.meta.effbandwidth
        del i1
    print("... they do!")


    # Exit here if the user requested test_only
    if args.test_only:
        return

    ### The Heavy Lifting ###
    # Run the spectral conversions
    if args.clear_all_spectral_files:
        print('')
        print("Deleting all spectral files...")
        for x in flist:
            i1 = geoio.dg.DGImage(x)
            i1.delete_all_spectral_files(test_only=False)
            del i1

    if args.spectral_ops:
        print('')
        print("Each file will be converted to {}:  ".format(args.spectral_ops))
        print('Running the spectral conversions...')
        for x in flist:
            print('')
            print("Staring on file {}:".format(x))
            i1 = geoio.dg.DGImage(x)
            if 'radiance' in args.spectral_ops:
                print("Running at sensor radiance...")
                i1.create_at_sensor_rad_files()
            if 'toa' in args.spectral_ops:
                print("Running toa reflectance...")
                i1.create_toa_ref_files()
            if 'DGAComp' in args.spectral_ops:
                print("Running DGAComp reflectance...")
                i1.create_dgacomp_ref_files()
            del i1

    print('*** Run Complete ***')


### Repeated string operation ###
def return_from_single_string(x):
    if os.path.isfile(x):
        # Then this is a file so just add it to the list
        return (x)
    else:
        # Then this is probably a glob pattern
        (path, base) = os.path.split(x)
        if not path:
            path = os.getcwd()
        return tt.files.search(path, base, depth=0)


if __name__ == "__main__":
    args = setup()
    main(args)



#!/usr/bin/env python

import argparse
import os
import tinytools as tt
import geoio


def setup():
    ### parse command line input ###
    descrip = """Create the spectral conversion files requested by the arguments."""

    parser = argparse.ArgumentParser(description=descrip)
    parser.add_argument("-input", "--input",
                        required=True,
                        nargs='+',
                        help="Input can be of four forms:"
                             "1. A list of file strings to operate on (this "
                                "will simply be passed through)."
                             "2. A string or list of strings with glob type "
                                "search patterns."
                             "3. A directory that will be searched for "
                                ".TIL files"
                             "4. A file to parse for a list of file strings")

    return parser.parse_args()

def main(args):
    ### Find files to run on ###
    # Figure out what the input format is and find files
    input = args.input
    flist = []

    # print(args.input)
    if len(input) > 1:
        # if isinstance(input,list):
        # Then this is a list of file strings or a list of glob patterns
        for x in input:
            # This decided if it is a file or a glob and expands as needed
            flist.extend(return_from_single_string(x))
    # elif isinstance(input,str):
    else:
        input = input[0]
        print(input)
        # If just a string, then it could be a file, glob, dir to search, or a
        # a file to read.
        if os.path.isdir(input):
            flist.extend(tt.files.search(input, '*.TIL', depth=0))
        elif os.path.isfile(input):
            e = os.path.splitext(input)[-1]
            if tt.files.filter(e, '.txt', case_sensitive=False):
                # Then this is a file of file strings
                print('*** NOT IMPLEMENTED YET ***')
        else:
            flist.extend(return_from_single_string(input))

    return flist

### Repeated string operation ###
def return_from_single_string(x):
    if os.path.isfile(x):
        # Then this is a file so just add it to the list
        return (x)
    else:
        # Then this is probably a glob pattern
        (path, base) = os.path.split(x)
        if not path:
            path = os.getcwd()
        return tt.files.search(path, base, depth=0)


if __name__ == "__main__":
    args = setup()
    main(args)



from setuptools import setup, find_packages

install_requires = [
    'gdal',
    'xmltodict',
    'pytz',
    'tzwhere',
    'ephem',
    'numpy',
    'tinytools'
    ]

tests_require = [
    'dgsamples'
]

setup(
    name='geoio',
    version='1.2.2',
    author='Nathan Longbotham',
    author_email='nlongbotham@digitalglobe.com',
    packages=find_packages(),
    description='Geo image reading/writing tools.',
    long_description=open('README.rst').read(),
    install_requires=install_requires,
    tests_require = tests_require,
    scripts=['bin/run_spectral_conversions.py','bin/find_files.py']
)
